"""
FIXME:
 - this is gross and temporary
  - move to global registry somehow
"""
import threading
import typing as ta

from .base.types import MarshalerFactory
from .base.types import UnmarshalerFactory
from .composite.iterables import IterableMarshalerFactory
from .composite.iterables import IterableUnmarshalerFactory
from .composite.literals import LiteralMarshalerFactory
from .composite.literals import LiteralUnmarshalerFactory
from .composite.mappings import MappingMarshalerFactory
from .composite.mappings import MappingUnmarshalerFactory
from .composite.maybes import MaybeMarshalerFactory
from .composite.maybes import MaybeUnmarshalerFactory
from .composite.newtypes import NewtypeMarshalerFactory
from .composite.newtypes import NewtypeUnmarshalerFactory
from .composite.optionals import OptionalMarshalerFactory
from .composite.optionals import OptionalUnmarshalerFactory
from .composite.special import SequenceNotStrMarshalerFactory
from .composite.special import SequenceNotStrUnmarshalerFactory
from .composite.unions.literals import LiteralUnionMarshalerFactory
from .composite.unions.literals import LiteralUnionUnmarshalerFactory
from .composite.unions.primitives import PrimitiveUnionMarshalerFactory
from .composite.unions.primitives import PrimitiveUnionUnmarshalerFactory
from .factories.invalidate import InvalidatableMarshalerFactory
from .factories.invalidate import InvalidatableUnmarshalerFactory
from .factories.moduleimport.factories import ModuleImportingMarshalerFactory
from .factories.moduleimport.factories import ModuleImportingUnmarshalerFactory
from .factories.multi import MultiMarshalerFactory
from .factories.multi import MultiUnmarshalerFactory
from .factories.recursive import RecursiveMarshalerFactory
from .factories.recursive import RecursiveUnmarshalerFactory
from .factories.typecache import TypeCacheMarshalerFactory
from .factories.typecache import TypeCacheUnmarshalerFactory
from .objects.dataclasses import DataclassMarshalerFactory
from .objects.dataclasses import DataclassUnmarshalerFactory
from .objects.namedtuples import NamedtupleMarshalerFactory
from .objects.namedtuples import NamedtupleUnmarshalerFactory
from .singular.datetimes import DATETIME_MARSHALER_FACTORY
from .singular.datetimes import DATETIME_UNMARSHALER_FACTORY
from .singular.enums import EnumMarshalerFactory
from .singular.enums import EnumUnmarshalerFactory
from .singular.numbers import NUMBERS_MARSHALER_FACTORY
from .singular.numbers import NUMBERS_UNMARSHALER_FACTORY
from .singular.primitives import PRIMITIVE_MARSHALER_FACTORY
from .singular.primitives import PRIMITIVE_UNMARSHALER_FACTORY
from .singular.uuids import UUID_MARSHALER_FACTORY
from .singular.uuids import UUID_UNMARSHALER_FACTORY
from .trivial.any import ANY_MARSHALER_FACTORY
from .trivial.any import ANY_UNMARSHALER_FACTORY


##


class StandardFactories(ta.NamedTuple):
    marshaler_factories: ta.Sequence[MarshalerFactory]
    unmarshaler_factories: ta.Sequence[UnmarshalerFactory]


DEFAULT_STANDARD_FACTORIES = StandardFactories(
    (
        PRIMITIVE_MARSHALER_FACTORY,
        NewtypeMarshalerFactory(),
        OptionalMarshalerFactory(),
        LiteralUnionMarshalerFactory(),
        PrimitiveUnionMarshalerFactory(),
        DataclassMarshalerFactory(),
        NamedtupleMarshalerFactory(),
        EnumMarshalerFactory(),
        LiteralMarshalerFactory(),
        NUMBERS_MARSHALER_FACTORY,
        UUID_MARSHALER_FACTORY,
        DATETIME_MARSHALER_FACTORY,
        MaybeMarshalerFactory(),
        MappingMarshalerFactory(),
        SequenceNotStrMarshalerFactory(),
        IterableMarshalerFactory(),
        ANY_MARSHALER_FACTORY,
    ),

    (
        PRIMITIVE_UNMARSHALER_FACTORY,
        NewtypeUnmarshalerFactory(),
        OptionalUnmarshalerFactory(),
        LiteralUnionUnmarshalerFactory(),
        PrimitiveUnionUnmarshalerFactory(),
        DataclassUnmarshalerFactory(),
        NamedtupleUnmarshalerFactory(),
        EnumUnmarshalerFactory(),
        LiteralUnmarshalerFactory(),
        NUMBERS_UNMARSHALER_FACTORY,
        UUID_UNMARSHALER_FACTORY,
        DATETIME_UNMARSHALER_FACTORY,
        MaybeUnmarshalerFactory(),
        MappingUnmarshalerFactory(),
        SequenceNotStrUnmarshalerFactory(),
        IterableUnmarshalerFactory(),
        ANY_UNMARSHALER_FACTORY,
    ),
)


##


def new_standard_marshaler_factory(
        *,
        first: ta.Iterable[MarshalerFactory] | None = None,
        last: ta.Iterable[MarshalerFactory] | None = None,
) -> MarshalerFactory:
    gl: ta.Any = None

    def fi_fn():
        nonlocal gl
        gl = DEFAULT_STANDARD_FACTORIES

        fi: MarshalerFactory = MultiMarshalerFactory(
            *(first if first is not None else []),
            *gl.marshaler_factories,
            *(last if last is not None else []),
        )

        fi = RecursiveMarshalerFactory(fi)

        fi = TypeCacheMarshalerFactory(fi)

        return fi

    fo: MarshalerFactory = (iv := InvalidatableMarshalerFactory(
        fi_fn,
        lambda: DEFAULT_STANDARD_FACTORIES is not gl,
    ))

    fo = ModuleImportingMarshalerFactory(fo, iv.invalidate)

    return fo


def new_standard_unmarshaler_factory(
        *,
        first: ta.Iterable[UnmarshalerFactory] | None = None,
        last: ta.Iterable[UnmarshalerFactory] | None = None,
) -> UnmarshalerFactory:
    gl: ta.Any = None

    def fi_fn():
        nonlocal gl
        gl = DEFAULT_STANDARD_FACTORIES

        fi: UnmarshalerFactory = MultiUnmarshalerFactory(
            *(first if first is not None else []),
            *gl.unmarshaler_factories,
            *(last if last is not None else []),
        )

        fi = RecursiveUnmarshalerFactory(fi)

        fi = TypeCacheUnmarshalerFactory(fi)

        return fi

    fo: UnmarshalerFactory = (iv := InvalidatableUnmarshalerFactory(
        fi_fn,
        lambda: DEFAULT_STANDARD_FACTORIES is not gl,
    ))

    fo = ModuleImportingUnmarshalerFactory(fo, iv.invalidate)

    return fo


##


_GLOBAL_LOCK = threading.RLock()


def install_standard_factories(
        *factories: MarshalerFactory | UnmarshalerFactory,
) -> None:
    with _GLOBAL_LOCK:
        global DEFAULT_STANDARD_FACTORIES

        m_lst: list[MarshalerFactory] = list(DEFAULT_STANDARD_FACTORIES.marshaler_factories)
        u_lst: list[UnmarshalerFactory] = list(DEFAULT_STANDARD_FACTORIES.unmarshaler_factories)

        for f in factories:
            k = False

            if isinstance(f, MarshalerFactory):
                m_lst[0:0] = [f]
                k = True

            if isinstance(f, UnmarshalerFactory):
                u_lst[0:0] = [f]
                k = True

            if not k:
                raise TypeError(f)

        new = StandardFactories(
            tuple(m_lst),
            tuple(u_lst),
        )

        if new != DEFAULT_STANDARD_FACTORIES:
            DEFAULT_STANDARD_FACTORIES = new



"""
TODO:
 - MarshalOption, UnmarshalOption, check if a context got something unexpected
 - Options only apply to Factories?
  - MarshalFactoryOption, MarshalOption, UnmarshalFactoryOption, MarshalOption ?
"""


class Option:
    pass



import dataclasses as dc
import json

from .plans import Plans


##


@dc.dataclass(frozen=True)
class DataclassTransformManifest:
    qualname: str
    digest: str
    plans: Plans

    def to_json(self) -> str:
        return json.dumps(
            dc.asdict(self),  # noqa
            indent=2,
        )



import re
import uuid

from ... import check
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..base.values import Value
from ..factories.typemap import TypeMapMarshalerFactory
from ..factories.typemap import TypeMapUnmarshalerFactory


##


PATTERN = re.compile(r'([0-9A-Fa-f]{8}-([0-9A-Fa-f]{4}-){3}[0-9A-Fa-f]{12})|([0-9A-Fa-f]{32})')


class UuidMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: uuid.UUID) -> Value:
        return str(o)

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> uuid.UUID:
        return uuid.UUID(check.isinstance(v, str).replace('-', ''))


UUID_MARSHALER_UNMARSHALER = UuidMarshalerUnmarshaler()

UUID_MARSHALER_FACTORY = TypeMapMarshalerFactory({uuid.UUID: UUID_MARSHALER_UNMARSHALER})
UUID_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({uuid.UUID: UUID_MARSHALER_UNMARSHALER})



import io

from ... import check
from .states import ACTIONS_IN_ORDER
from .states import STATE_TABLES
from .states import STATES
from .states import STATES_IN_ORDER
from .states import StateTransition
from .states import check_state_tables


##


def _pad(s: str, length: int) -> str:
    return s + ' ' * (length - len(s))


def _generate_c() -> dict[str, str]:
    out = {}

    #

    f = io.StringIO()

    f.write('typedef enum {\n')
    for i, state in enumerate(STATES_IN_ORDER):
        f.write(f'   VTPARSE_STATE_{state.upper()} = {i + 1},\n')
    f.write('} vtparse_state_t;\n\n')

    f.write('typedef enum {\n')
    for i, action_str in enumerate(ACTIONS_IN_ORDER):
        f.write(f'   VTPARSE_ACTION_{action_str.upper()} = {i + 1},\n')
    f.write('} vtparse_action_t;\n\n')

    f.write('typedef unsigned char state_change_t;\n')
    f.write(f'extern state_change_t STATE_TABLE[{len(STATES_IN_ORDER)}][256];\n')
    f.write(f'extern vtparse_action_t ENTRY_ACTIONS[{len(STATES_IN_ORDER)}];\n')
    f.write(f'extern vtparse_action_t EXIT_ACTIONS[{len(STATES_IN_ORDER)}];\n')
    f.write(f'extern char *ACTION_NAMES[{len(ACTIONS_IN_ORDER) + 1}];\n')
    f.write(f'extern char *STATE_NAMES[{len(STATES_IN_ORDER) + 1}];\n\n')

    out['vtparse_table.h'] = f.getvalue()

    #

    f = io.StringIO()

    f.write('#include "vtparse_table.h"\n\n')

    f.write('char *ACTION_NAMES[] = {\n')
    f.write('   "<no action>",\n')
    for action_str in ACTIONS_IN_ORDER:
        f.write(f'   "{action_str.upper()}",\n')
    f.write('};\n\n')

    f.write('char *STATE_NAMES[] = {\n')
    f.write('   "<no state>",\n')
    for state in STATES_IN_ORDER:
        f.write(f'   "{state}",\n')
    f.write('};\n\n')

    f.write(f'state_change_t STATE_TABLE[{len(STATES_IN_ORDER)}][256] = {{\n')
    for i, state in enumerate(STATES_IN_ORDER):
        f.write(f'  {{  /* VTPARSE_STATE_{state.upper()} = {i} */\n')
        for j, state_change in enumerate(STATE_TABLES[state]):
            if not state_change:
                f.write('    0,\n')
            else:
                action = next((s for s in state_change if isinstance(s, str)), None)
                state_trans = next((s for s in state_change if isinstance(s, StateTransition)), None)
                action_str = f'VTPARSE_ACTION_{action.upper()}' if action else '0'
                state_str = f'VTPARSE_STATE_{state_trans.to_state}' if state_trans else '0'
                f.write(
                    f'/*{str(j).rjust(3)}*/  {_pad(action_str, 33)} | ({_pad(state_str, 33)} << 4),\n')
        f.write('  },\n')
    f.write('};\n\n')

    f.write('vtparse_action_t ENTRY_ACTIONS[] = {\n')
    for state in STATES_IN_ORDER:
        actions = STATES[state]
        if 'on_entry' in actions:
            f.write(f"   VTPARSE_ACTION_{check.isinstance(actions['on_entry'], str).upper()}, /* {state} */\n")
        else:
            f.write(f'   0  /* none for {state} */,\n')
    f.write('};\n\n')

    f.write('vtparse_action_t EXIT_ACTIONS[] = {\n')
    for state in STATES_IN_ORDER:
        actions = STATES[state]
        if 'on_exit' in actions:
            f.write(f"   VTPARSE_ACTION_{check.isinstance(actions['on_exit'], str).upper()}, /* {state} */\n")
        else:
            f.write(f'   0  /* none for {state} */,\n')
    f.write('};\n\n')

    out['vtparse_table.c'] = f.getvalue()

    #

    return out


if __name__ == '__main__':
    check_state_tables(STATE_TABLES)
    for f, c in _generate_c().items():
        print(f)
        print(c)
        print()



"""
TODO:
 - merged compose configs: https://github.com/wrmsr/bane/blob/27647abdcfb323b73e6982a5c318c7029496b203/core/dev/docker/compose.go#L38
"""  # noqa
import typing as ta

from .. import check
from .. import lang


if ta.TYPE_CHECKING:
    import yaml
else:
    yaml = lang.proxy_import('yaml')


##


class ComposeConfig:
    def __init__(
            self,
            prefix: str,
            *,
            file_path: str | None = None,
    ) -> None:
        super().__init__()

        self._prefix = prefix
        self._file_path = file_path

    @lang.cached_function
    def get_config(self) -> ta.Mapping[str, ta.Any]:
        with open(check.not_none(self._file_path)) as f:
            buf = f.read()
        return yaml.safe_load(buf)

    @lang.cached_function
    def get_services(self) -> ta.Mapping[str, ta.Any]:
        ret = {}
        for n, c in self.get_config()['services'].items():
            check.state(n.startswith(self._prefix))
            ret[n[len(self._prefix):]] = c

        return ret


def get_compose_port(cfg: ta.Mapping[str, ta.Any], default: int) -> int:
    return check.single(
        int(l)
        for p in cfg['ports']
        for l, r in [p.split(':')]
        if int(r) == default
    )



import typing as ta


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class ScopedChainDict(ta.Generic[K, V]):
    """
    Dictionary that can delegate lookups to multiple dicts. This provides a basic get/set dict interface that is backed
    by multiple dicts.  Each dict is searched from the top most (most recently pushed) scope dict until a match is
    found.
    """

    def __init__(self, *scopes: ta.Mapping[K, V]) -> None:
        super().__init__()

        # The scopes are evaluated starting at the top of the stack (the most recently pushed scope via .push_scope()).
        # If we use a normal list() and push/pop scopes by adding/removing to the end of the list, we'd have to always
        # call reversed(self._scopes) whenever we resolve a key, because the end of the list is the top of the stack.
        # To avoid this, we're using a deque so we can append to the front of the list via .appendleft() in constant
        # time, and iterate over scopes without having to do so with a reversed() call each time.
        self._scopes: list[ta.Mapping[K, V]] = list(reversed(scopes))

    def __getitem__(self, key: K) -> V:
        for scope in reversed(self._scopes):
            if key in scope:
                return scope[key]
        raise KeyError(key)

    def get(self, key: K, default: V | None = None) -> V | None:
        try:
            return self[key]
        except KeyError:
            return default

    def push_scope(self, scope: ta.Mapping[K, V]) -> None:
        self._scopes.append(scope)

    def pop_scope(self) -> None:
        self._scopes.pop()



import dataclasses as dc
import sys
import typing as ta

from ..... import check
from ..... import lang
from ...._internals import StdFieldType
from ...._internals import std_field_type
from ....debug import DEBUG
from ....specs import DefaultFactory
from ....specs import FieldSpec
from ....specs import FieldType
from .metadata import _ExtraFieldParamsMetadata
from .metadata import set_field_spec_metadata


##


_IS_PY_3_14 = sys.version_info >= (3, 14)


##


STD_FIELD_TYPE_BY_SPEC_FIELD_TYPE: ta.Mapping[FieldType, StdFieldType] = {
    FieldType.INSTANCE: StdFieldType.INSTANCE,
    FieldType.CLASS_VAR: StdFieldType.CLASS_VAR,
    FieldType.INIT_VAR: StdFieldType.INIT_VAR,
}

SPEC_FIELD_TYPE_BY_STD_FIELD_TYPE = {v: k for k, v in STD_FIELD_TYPE_BY_SPEC_FIELD_TYPE.items()}


#


def std_to_spec_field_default(
        *,
        default: ta.Any,
        default_factory: ta.Any,
) -> lang.Maybe[ta.Any]:
    if default is not dc.MISSING:
        check.state(default_factory is dc.MISSING)
        return lang.just(default)
    elif default_factory is not dc.MISSING:
        return lang.just(DefaultFactory(default_factory))
    else:
        return lang.empty()


def std_field_to_spec_field_default(f: dc.Field) -> lang.Maybe[ta.Any]:
    return std_to_spec_field_default(
        default=f.default,
        default_factory=f.default_factory,
    )


class StdDefaults(ta.NamedTuple):
    default: ta.Any
    default_factory: ta.Any


def spec_field_default_to_std_defaults(dfl: lang.Maybe[DefaultFactory | ta.Any]) -> StdDefaults:
    if not dfl.present:
        return StdDefaults(dc.MISSING, dc.MISSING)
    elif isinstance(dfv := dfl.must(), DefaultFactory):
        return StdDefaults(dc.MISSING, dfv.fn)
    else:
        return StdDefaults(dfv, dc.MISSING)


#


def std_field_to_field_spec(
        f: dc.Field,
        *,
        ignore_metadata: bool = False,
        ignore_extra_params: bool = False,
        set_metadata: bool = False,
) -> FieldSpec:
    if not ignore_metadata:
        try:
            fs = f.metadata[FieldSpec]
        except KeyError:
            pass
        else:
            check_field_spec_against_field(f, fs)
            return fs

    extra_params = {}
    if not ignore_extra_params:
        extra_params.update(f.metadata.get(_ExtraFieldParamsMetadata, {}))

    fs = FieldSpec(
        name=check.non_empty_str(f.name),
        annotation=check.is_not(f.type, dc.MISSING),

        default=std_field_to_spec_field_default(f),

        init=check.isinstance(f.init, bool) if DEBUG else f.init,
        repr=check.isinstance(f.repr, bool) if DEBUG else f.repr,
        hash=check.isinstance(f.hash, (bool, None)) if DEBUG else f.hash,
        compare=check.isinstance(f.compare, bool) if DEBUG else f.compare,
        metadata=f.metadata,
        kw_only=None if f.kw_only is dc.MISSING else (check.isinstance(f.kw_only, bool) if DEBUG else f.kw_only),

        **lang.opt_kw(
            doc=extra_params.get('doc', f.doc if _IS_PY_3_14 else None),  # type: ignore[attr-defined]  # noqa
        ),

        **lang.opt_kw(
            coerce=extra_params.get('coerce'),
            validate=extra_params.get('validate'),
            check_type=extra_params.get('check_type'),
            override=extra_params.get('override'),
            repr_fn=extra_params.get('repr_fn'),
            repr_priority=extra_params.get('repr_priority'),
        ),

        field_type=SPEC_FIELD_TYPE_BY_STD_FIELD_TYPE[std_field_type(f)],
    )

    if set_metadata:
        set_field_spec_metadata(f, fs)

    return fs


##


def field_spec_to_std_field(fs: FieldSpec) -> dc.Field:
    sdf = spec_field_default_to_std_defaults(fs.default)

    f = dc.Field(
        default=sdf.default,
        default_factory=sdf.default_factory,

        init=fs.init,
        repr=fs.repr,
        hash=fs.hash,
        compare=fs.compare,
        **lang.opt_kw(metadata=fs.metadata),
        kw_only=dc.MISSING if fs.kw_only is None else fs.kw_only,  # type: ignore[arg-type]
        **(dict(doc=fs.doc) if _IS_PY_3_14 else {}),
    )

    f.name = fs.name
    f.type = fs.annotation

    f._field_type = STD_FIELD_TYPE_BY_SPEC_FIELD_TYPE[fs.field_type].value  # type: ignore[attr-defined]  # noqa

    set_field_spec_metadata(f, fs)

    return f


##


def check_field_spec_against_field(f: dc.Field, fs: FieldSpec) -> None:
    f_dct = {
        'name': f.name,
        'type': f.type,

        'default': f.default,
        'default_factory': f.default_factory,

        'repr': f.repr,
        'hash': f.hash,
        'init': f.init,
        'compare': f.compare,
        # f.metadata,
        'kw_only': f.kw_only if f.kw_only is not dc.MISSING else None,
        **({'doc': f.doc} if _IS_PY_3_14 else {}),  # type: ignore[attr-defined]  # noqa

        'std_field_type': f._field_type,  # type: ignore[attr-defined]  # noqa
    }

    fs_dct = {
        'name': fs.name,
        'type': fs.annotation,

        **spec_field_default_to_std_defaults(fs.default)._asdict(),

        'repr': fs.repr,
        'hash': fs.hash,
        'init': fs.init,
        'compare': fs.compare,
        # fs.metadata,
        'kw_only': fs.kw_only,
        **({'doc': fs.doc} if _IS_PY_3_14 else {}),

        'std_field_type': STD_FIELD_TYPE_BY_SPEC_FIELD_TYPE[fs.field_type].value,
    }

    if f_dct != fs_dct:
        diff_dct = {
            k: (f_v, fs_v)
            for k, f_v in f_dct.items()
            if (fs_v := fs_dct[k]) != f_v
        }
        raise RuntimeError(f'Field/FieldSpec mismatch: {diff_dct!r}')



import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import reflect as rfl
from ..base.contexts import MarshalFactoryContext
from ..base.contexts import UnmarshalFactoryContext
from ..base.types import Marshaler
from ..base.types import MarshalerFactory
from ..base.types import Unmarshaler
from ..base.types import UnmarshalerFactory


##


@dc.dataclass(frozen=True)
class TypeMapMarshalerFactory(MarshalerFactory):
    m: ta.Mapping[rfl.Type, Marshaler | MarshalerFactory] = dc.xfield(
        default_factory=dict,
        coerce=lambda d: {
            check.isinstance(k, rfl.TYPES): check.isinstance(v, (Marshaler, UnmarshalerFactory))
            for k, v in d.items()
        },
    )

    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        check.isinstance(rty, rfl.TYPES)
        try:
            m = self.m[rty]
        except KeyError:
            return None

        if isinstance(m, Marshaler):
            return lambda: m
        elif isinstance(m, MarshalerFactory):
            return m.make_marshaler(ctx, rty)
        else:
            raise TypeError(m)


@dc.dataclass(frozen=True)
class TypeMapUnmarshalerFactory(UnmarshalerFactory):
    u: ta.Mapping[rfl.Type, Unmarshaler | UnmarshalerFactory] = dc.xfield(
        default_factory=dict,
        coerce=lambda d: {
            check.isinstance(k, rfl.TYPES): check.isinstance(v, (Unmarshaler, UnmarshalerFactory))
            for k, v in d.items()
        },
    )

    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        check.isinstance(rty, rfl.TYPES)
        try:
            u = self.u[rty]
        except KeyError:
            return None

        if isinstance(u, Unmarshaler):
            return lambda: u
        elif isinstance(u, UnmarshalerFactory):
            return u.make_unmarshaler(ctx, rty)
        else:
            raise TypeError(u)



import threading
import typing as ta

from ... import lang
from ._compat import compat
from .app import App
from .cvs import Cvs
from .requests import Request
from .responses import Response


##


class Api:
    def __init__(
            self,
            *,
            base_app_cls: type[App] = App,
    ) -> None:
        super().__init__()

        self._base_app_cls = base_app_cls

        self._lock = threading.RLock()

    @property
    def _cls_name_repr(self) -> str:
        return f'{self.__class__.__name__}@{id(self):x}'

    def __repr__(self) -> str:
        return f'{self._cls_name_repr}()'

    ##
    # app cls

    _app_cls: type[App]

    @property
    def app_cls(self) -> type[App]:
        try:
            return self._app_cls
        except AttributeError:
            pass

        with self._lock:
            try:
                return self._app_cls
            except AttributeError:
                pass

            self._app_cls = lang.new_type(  # noqa
                f'{self._base_app_cls.__name__}<{self._cls_name_repr}>',
                (self._base_app_cls,),
                {'_api': self},
            )

            return self._app_cls

    @property
    @compat
    def Flask(self) -> type[App]:  # noqa
        return self.app_cls

    ##
    # helpers

    def abort(self, code: int | Response, *args: ta.Any, **kwargs: ta.Any) -> ta.NoReturn:
        raise NotImplementedError

    ##
    # cv's

    @property
    @compat
    def request(self) -> Request:
        return Cvs.REQUEST.get()

    ##
    # type aliases - must be last

    Response = Response



from .... import check
from ..processing.base import Processor
from ..processing.priority import ProcessorPriority
from ..processing.registry import register_processor_type
from ..utils import set_new_attribute
from .fields import InitFields


##


@register_processor_type(priority=ProcessorPriority.POST_GENERATION)
class MatchArgsProcessor(Processor):
    def check(self) -> None:
        check.not_none(self._ctx[InitFields])

    def process(self, cls: type) -> type:
        if not self._ctx.cs.match_args or '__match_args__' in self._ctx.cls.__dict__:
            return cls

        set_new_attribute(
            cls,
            '__match_args__',
            tuple(f.name for f in self._ctx[InitFields].std),
        )

        return cls



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import typing as ta

from .types import ABSENT_TYPED_LOGGER_VALUE
from .types import AbsentTypedLoggerValue
from .types import ResolvedTypedLoggerFieldValue
from .types import TypedLoggerFieldValue
from .types import TypedLoggerValue
from .types import TypedLoggerValueOrAbsent
from .types import UnwrappedTypedLoggerFieldValue
from .types import unwrap_typed_logger_field_value


if ta.TYPE_CHECKING:
    from .bindings import TypedLoggerBindings


TypedLoggerValueT = ta.TypeVar('TypedLoggerValueT', bound='TypedLoggerValue')


##


class RecursiveTypedLoggerValueError(Exception):
    def __init__(self, cls: ta.Type[TypedLoggerValue], rec: ta.Sequence[ta.Type[TypedLoggerValue]]) -> None:
        super().__init__()

        self.cls = cls
        self.rec = rec

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}('
            f'cls={self.cls!r}, '
            f'rec={self.rec!r}'
            f')'
        )


class UnboundTypedLoggerValueError(Exception):
    def __init__(self, cls: ta.Type[TypedLoggerValue]) -> None:
        super().__init__()

        self.cls = cls

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(cls={self.cls!r})'


#


@ta.final
class TypedLoggerContext:
    def __init__(
            self,
            bindings: 'TypedLoggerBindings',
            *,
            no_auto_default_values: bool = False,
            default_absent: bool = False,
    ) -> None:
        super().__init__()

        self._bindings = bindings
        self._no_auto_default_values = no_auto_default_values
        self._default_absent = default_absent

        self._values: ta.Dict[ta.Type[TypedLoggerValue], TypedLoggerValueOrAbsent] = dict(bindings.const_value_map)  # noqa
        self._rec: ta.Dict[ta.Type[TypedLoggerValue], None] = {}

    @property
    def bindings(self) -> 'TypedLoggerBindings':
        return self._bindings

    @property
    def no_auto_default_values(self) -> bool:
        return self._no_auto_default_values

    @property
    def default_absent(self) -> bool:
        return self._default_absent

    #

    def __getitem__(self, cls: ta.Type[TypedLoggerValueT]) -> ta.Union[TypedLoggerValueT, AbsentTypedLoggerValue]:
        try:
            return self._values[cls]  # type: ignore[return-value]
        except KeyError:
            pass

        if not issubclass(cls, TypedLoggerValue):
            raise TypeError(cls)

        if cls in self._rec:
            raise RecursiveTypedLoggerValueError(cls, list(self._rec))

        self._rec[cls] = None
        try:
            v: ta.Union[TypedLoggerValueOrAbsent]

            if (bv := self._bindings.lookup_value(cls)) is not None:
                if bv is ABSENT_TYPED_LOGGER_VALUE:  # noqa
                    v = ABSENT_TYPED_LOGGER_VALUE

                else:
                    v = bv._typed_logger_provide_value(self)  # noqa

            else:
                if not self._no_auto_default_values and (dt := cls._typed_logger_maybe_provide_default_value(self)):
                    [v] = dt

                elif self._default_absent:
                    v = ABSENT_TYPED_LOGGER_VALUE

                else:
                    raise UnboundTypedLoggerValueError(cls) from None

            self._values[cls] = v
            return v  # type: ignore[return-value]

        finally:
            self._rec.pop(cls)

    #

    def resolve_field_value(self, fv: TypedLoggerFieldValue) -> ResolvedTypedLoggerFieldValue:
        if fv is ABSENT_TYPED_LOGGER_VALUE:
            return fv  # type: ignore[return-value]

        elif isinstance(fv, type):
            return self[fv]._typed_logger_resolve_field_value(self)  # type: ignore[type-var]  # noqa

        else:
            return fv._typed_logger_resolve_field_value(self)  # noqa

    def unwrap_field_value(self, fv: TypedLoggerFieldValue) -> UnwrappedTypedLoggerFieldValue:
        return unwrap_typed_logger_field_value(self.resolve_field_value(fv))



import contextlib
import dataclasses as dc
import functools
import threading
import typing as ta


T = ta.TypeVar('T')
P = ta.ParamSpec('P')


##


_LOCK = threading.RLock()
_LOCAL: threading.local


def _local() -> threading.local:
    global _LOCAL

    try:
        return _LOCAL
    except NameError:
        pass

    with _LOCK:
        try:
            return _LOCAL
        except NameError:
            pass

        _LOCAL = threading.local()
        return _LOCAL


def _depth_map() -> dict[ta.Any, int]:
    lo = _local()
    try:
        return lo.depth_map
    except AttributeError:
        dm = lo.depth_map = {}
        return dm


##


@dc.dataclass()
class LimitedRecursionError(RecursionError):
    key: ta.Any
    depth: int


@contextlib.contextmanager
def recursion_limiting_context(key: ta.Any, limit: int | None) -> ta.Iterator[int | None]:
    if limit is None:
        yield None
        return

    dm = _depth_map()

    try:
        pd: int | None = dm[key]
    except KeyError:
        pd = None
    else:
        if not isinstance(pd, int) and pd > 0:  # type: ignore[operator]
            raise RuntimeError

    if pd is not None and pd >= limit:
        raise LimitedRecursionError(key, pd)

    nd = (pd or 0) + 1
    dm[key] = nd

    try:
        yield nd

    finally:
        if dm.get(key) != nd:
            raise RuntimeError

        if pd is not None:
            dm[key] = pd
        else:
            del dm[key]


##


def recursion_limiting(limit: int | None) -> ta.Callable[[ta.Callable[P, T]], ta.Callable[P, T]]:
    def outer(fn):
        if not callable(fn):
            raise TypeError(fn)

        if limit is None:
            return fn

        @functools.wraps(fn)
        def inner(*args, **kwargs):
            with recursion_limiting_context(fn, limit):
                return fn(*args, **kwargs)

        return inner

    return outer



import typing as ta

from .... import check
from .... import collections as col
from .... import dataclasses as dc
from .... import lang
from .... import reflect as rfl
from ...base.contexts import MarshalContext
from ...base.contexts import MarshalFactoryContext
from ...base.contexts import UnmarshalContext
from ...base.contexts import UnmarshalFactoryContext
from ...base.types import Marshaler
from ...base.types import MarshalerFactory
from ...base.types import Unmarshaler
from ...base.types import UnmarshalerFactory
from ...base.values import Value


##


LITERAL_UNION_TYPES: tuple[type, ...] = (
    int,
    str,
)


class DestructuredLiteralUnionType(ta.NamedTuple):
    v_ty: type
    lit: rfl.Literal
    non_lit: rfl.Type


def _destructure_literal_union_type(rty: rfl.Type) -> DestructuredLiteralUnionType | None:
    if not isinstance(rty, rfl.Union):
        return None
    lits, non_lits = col.partition(rty.args, lang.isinstance_of(rfl.Literal))  # noqa
    if len(lits) != 1 or len(non_lits) != 1:
        return None
    lit = check.isinstance(check.single(lits), rfl.Literal)
    v_tys = set(map(type, lit.args))
    if len(v_tys) != 1:
        return None
    [v_ty] = v_tys
    if v_ty in rty.args or v_ty not in LITERAL_UNION_TYPES:
        return None
    return DestructuredLiteralUnionType(v_ty, lit, check.single(non_lits))


#


@dc.dataclass(frozen=True)
class LiteralUnionMarshaler(Marshaler):
    v_ty: type
    l: Marshaler
    x: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        if isinstance(o, self.v_ty):
            return self.l.marshal(ctx, o)
        else:
            return self.x.marshal(ctx, o)


class LiteralUnionMarshalerFactory(MarshalerFactory):
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if (ds := _destructure_literal_union_type(rty)) is None:
            return None
        return lambda: LiteralUnionMarshaler(ds.v_ty, ctx.make_marshaler(ds.lit), ctx.make_marshaler(ds.non_lit))


#


@dc.dataclass(frozen=True)
class LiteralUnionUnmarshaler(Unmarshaler):
    v_ty: type
    l: Unmarshaler
    x: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        if isinstance(v, self.v_ty):
            return self.l.unmarshal(ctx, v)  # type: ignore[arg-type]
        else:
            return self.x.unmarshal(ctx, v)


class LiteralUnionUnmarshalerFactory(UnmarshalerFactory):
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if (ds := _destructure_literal_union_type(rty)) is None:
            return None
        return lambda: LiteralUnionUnmarshaler(ds.v_ty, ctx.make_unmarshaler(ds.lit), ctx.make_unmarshaler(ds.non_lit))



# ruff: noqa: UP006 UP007 UP045
import io
import types
import typing as ta

from .io import BufferedIOBase_AsyncIoProxy
from .io import BufferedRandom_AsyncIoProxy
from .io import BufferedReader_AsyncIoProxy
from .io import BufferedRWPair_AsyncIoProxy
from .io import BufferedWriter_AsyncIoProxy
from .io import BytesIO_AsyncIoProxy
from .io import FileIO_AsyncIoProxy
from .io import IOBase_AsyncIoProxy
from .io import RawIOBase_AsyncIoProxy
from .io import StringIO_AsyncIoProxy
from .io import TextIOBase_AsyncIoProxy
from .io import TextIOWrapper_AsyncIoProxy
from .proxy import AsyncIoProxyRunner
from .proxy import AsyncIoProxyTarget
from .proxy import async_io_proxy_cls_for
from .proxy import async_io_proxy_fn
from .typing import TypingBinaryIO_AsyncIoProxy
from .typing import TypingIO_AsyncIoProxy
from .typing import TypingTextIO_AsyncIoProxy


##


@ta.final
class AsyncIoProxier:
    def __init__(self, runner_policy: ta.Callable[[ta.Any], AsyncIoProxyRunner]) -> None:
        super().__init__()

        self._runner_policy = runner_policy

    def get_runner(self, obj: ta.Any) -> AsyncIoProxyRunner:
        return self._runner_policy(obj)

    ##

    def target_obj(self, obj: ta.Any) -> AsyncIoProxyTarget:
        runner = self.get_runner(obj)
        return AsyncIoProxyTarget(obj, runner)

    def proxy_obj_with_cls(self, obj, proxy_cls):
        target = self.target_obj(obj)
        proxy = proxy_cls(target)
        return proxy

    def maybe_proxy_obj(self, obj):
        if (proxy_cls := async_io_proxy_cls_for(obj)) is None:
            return obj
        return self.proxy_obj_with_cls(obj, proxy_cls)

    ##

    @ta.overload
    def proxy_obj(self, obj: io.StringIO) -> StringIO_AsyncIoProxy:  # type: ignore[overload-overlap]  # 1
        ...

    @ta.overload
    def proxy_obj(self, obj: io.TextIOWrapper) -> TextIOWrapper_AsyncIoProxy:  # type: ignore[overload-overlap]  # 2
        ...

    @ta.overload
    def proxy_obj(self, obj: io.TextIOBase) -> TextIOBase_AsyncIoProxy:  # 3
        ...

    @ta.overload
    def proxy_obj(self, obj: io.FileIO) -> FileIO_AsyncIoProxy:  # type: ignore[overload-overlap]  # 4
        ...

    @ta.overload
    def proxy_obj(self, obj: io.BufferedRandom) -> BufferedRandom_AsyncIoProxy:  # type: ignore[overload-overlap]  # 5
        ...

    @ta.overload
    def proxy_obj(self, obj: io.BufferedRWPair) -> BufferedRWPair_AsyncIoProxy:  # 6
        ...

    @ta.overload
    def proxy_obj(self, obj: io.BufferedWriter) -> BufferedWriter_AsyncIoProxy:  # type: ignore[overload-overlap]  # 7
        ...

    @ta.overload
    def proxy_obj(self, obj: io.BufferedReader) -> BufferedReader_AsyncIoProxy:  # type: ignore[overload-overlap]  # 8
        ...

    @ta.overload
    def proxy_obj(self, obj: io.BytesIO) -> BytesIO_AsyncIoProxy:  # type: ignore[overload-overlap]  # 9
        ...

    @ta.overload
    def proxy_obj(self, obj: io.BufferedIOBase) -> BufferedIOBase_AsyncIoProxy:  # 10
        ...

    @ta.overload
    def proxy_obj(self, obj: io.RawIOBase) -> RawIOBase_AsyncIoProxy:  # 11
        ...

    @ta.overload
    def proxy_obj(self, obj: io.IOBase) -> IOBase_AsyncIoProxy:  # 12
        ...

    #

    @ta.overload
    def proxy_obj(self, obj: ta.TextIO) -> TypingTextIO_AsyncIoProxy:  # 13
        ...

    @ta.overload
    def proxy_obj(self, obj: ta.BinaryIO) -> TypingBinaryIO_AsyncIoProxy:  # 14
        ...

    @ta.overload
    def proxy_obj(self, obj: ta.IO) -> TypingIO_AsyncIoProxy:  # 15
        ...

    #

    def proxy_obj(self, obj):
        if (proxy_cls := async_io_proxy_cls_for(obj)) is None:
            raise TypeError(obj)
        return self.proxy_obj_with_cls(obj, proxy_cls)

    ##

    def proxy_fn(self, fn, *, wrap_result='auto'):
        if wrap_result == 'auto':
            result_wrapper = self.maybe_proxy_obj
        elif wrap_result is True:
            result_wrapper = self.proxy_obj
        elif wrap_result is False:
            result_wrapper = None
        else:
            raise TypeError(wrap_result)
        runner = self.get_runner(fn)
        return async_io_proxy_fn(fn, runner, result_wrapper=result_wrapper)

    ##

    FN_TYPES: ta.Tuple[type, ...] = (
        types.BuiltinFunctionType,
        types.BuiltinMethodType,
        types.FunctionType,
        types.MethodType,
    )

    def proxy(self, obj):
        if isinstance(obj, self.FN_TYPES):
            return self.proxy_fn(obj)
        else:
            return self.proxy_obj(obj)



import typing as ta

from .... import check
from .... import collections as col
from .... import dataclasses as dc
from .... import reflect as rfl
from ...base.contexts import MarshalContext
from ...base.contexts import MarshalFactoryContext
from ...base.contexts import UnmarshalContext
from ...base.contexts import UnmarshalFactoryContext
from ...base.types import Marshaler
from ...base.types import MarshalerFactory
from ...base.types import Unmarshaler
from ...base.types import UnmarshalerFactory
from ...base.values import Value


##


PRIMITIVE_UNION_TYPES: tuple[type, ...] = (
    float,
    int,
    str,
    bool,
)


class DestructuredPrimitiveUnionType(ta.NamedTuple):
    prim: ta.Sequence[type]
    non_prim: ta.Sequence[rfl.Type]


def _destructure_primitive_union_type(
        rty: rfl.Type,
        prim_tys: ta.Container[type] = PRIMITIVE_UNION_TYPES,
) -> DestructuredPrimitiveUnionType | None:
    if not isinstance(rty, rfl.Union):
        return None

    pr = col.partition(rty.args, lambda a: isinstance(a, type) and a in prim_tys)
    if not pr.t or len(pr.f) > 1:
        return None

    return DestructuredPrimitiveUnionType(*pr)  # type: ignore[arg-type]


#


@dc.dataclass(frozen=True)
class PrimitiveUnionMarshaler(Marshaler):
    tys: ta.Sequence[type]
    x: Marshaler | None = None

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        if type(o) not in self.tys:
            if self.x is not None:
                return self.x.marshal(ctx, o)
            raise TypeError(o)
        return o


@dc.dataclass(frozen=True)
class PrimitiveUnionMarshalerFactory(MarshalerFactory):
    tys: ta.Sequence[type] = PRIMITIVE_UNION_TYPES

    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if (ds := _destructure_primitive_union_type(rty, self.tys)) is None:
            return None
        return lambda: PrimitiveUnionMarshaler(
            ds.prim,
            ctx.make_marshaler(check.single(ds.non_prim)) if ds.non_prim else None,
        )


#


@dc.dataclass(frozen=True)
class PrimitiveUnionUnmarshaler(Unmarshaler):
    tys: ta.Sequence[type]
    x: Unmarshaler | None = None

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        if type(v) not in self.tys:
            if self.x is not None:
                return self.x.unmarshal(ctx, v)
            raise TypeError(v)
        return v


@dc.dataclass(frozen=True)
class PrimitiveUnionUnmarshalerFactory(UnmarshalerFactory):
    tys: ta.Sequence[type] = PRIMITIVE_UNION_TYPES

    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if (ds := _destructure_primitive_union_type(rty, self.tys)) is None:
            return None
        return lambda: PrimitiveUnionUnmarshaler(
            ds.prim,
            ctx.make_unmarshaler(check.single(ds.non_prim)) if ds.non_prim else None,
        )



import abc
import functools
import typing as ta

from ..lite.abstract import Abstract
from ..lite.asyncs import sync_await as _sync_await
from ..lite.maybes import Maybe


O = ta.TypeVar('O')
I = ta.TypeVar('I')
R = ta.TypeVar('R')
T = ta.TypeVar('T')
P = ta.ParamSpec('P')


##


def as_async(fn: ta.Callable[P, T], *, wrap: bool = False) -> ta.Callable[P, ta.Awaitable[T]]:
    async def inner(*args, **kwargs):
        return fn(*args, **kwargs)

    return functools.wraps(fn)(inner) if wrap else inner


##


class AsyncGeneratorWithReturn(ta.AsyncGenerator[O, I], Abstract, ta.Generic[O, I, R]):
    @property
    @abc.abstractmethod
    def value(self) -> Maybe[R]:
        raise NotImplementedError


@ta.final
class _AsyncGeneratorWithReturn(AsyncGeneratorWithReturn[O, I, R]):
    def __init__(self, ag: ta.AsyncGenerator[O, I]) -> None:
        if isinstance(ag, _AsyncGeneratorWithReturn):
            raise TypeError(ag)
        self._ag = ag

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._ag!r})'

    _v: Maybe[R] = Maybe.empty()

    @property
    def value(self) -> Maybe[R]:
        return self._v

    def _set_value(self, v: R) -> None:
        if self._v.present:
            raise TypeError('Return value already set')
        self._v = Maybe.just(v)

    def __anext__(self):
        return self._ag.__anext__()

    def asend(self, value):
        return self._ag.asend(value)

    def athrow(self, typ, val=None, tb=None):
        return self._ag.athrow(typ, val, tb)

    def aclose(self):
        return self._ag.aclose()


def async_generator_with_return(
        fn: ta.Callable[ta.Concatenate[ta.Callable[[R], None], P], ta.AsyncGenerator[O, I]],
) -> ta.Callable[P, AsyncGeneratorWithReturn[O, I, R]]:
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        def set_value(v):
            x._set_value(v)  # noqa

        return (x := _AsyncGeneratorWithReturn(fn(set_value, *args, **kwargs)))  # type: ignore[var-annotated]

    return inner


##


sync_await = _sync_await


try:
    from . import _asyncs as cext  # type: ignore

except ImportError:
    pass

else:
    sync_await = cext.sync_await



"""
TODO:
 - reserver / generator

==

@dc.dataclass()
class _ReservedFilenameEntry:
    unique_id: str
    seq: int = 0


_RESERVED_FILENAME_UUID_THREAD_LOCAL = threading.local()


def reserve_filename(prefix: str) -> str:
    try:
        e = _RESERVED_FILENAME_UUID_THREAD_LOCAL.unique_id
    except AttributeError:
        e = _RESERVED_FILENAME_UUID_THREAD_LOCAL.unique_id = _ReservedFilenameEntry(str(uuid.uuid4()))
    while True:
        unique_filename = f'<generated:{prefix}:{e.seq}>'
        cache_line = (1, None, (e.unique_id,), unique_filename)
        e.seq += 1
        if linecache.cache.setdefault(unique_filename, cache_line) == cache_line:
            return unique_filename
"""
import typing as ta


##


class LinecacheKey(ta.NamedTuple):
    size: int  # os.stat().st_size
    mtime: float | None  # os.stat().st_mtime
    lines: ta.Sequence[str]
    fullname: str


LinecacheLazyLoader: ta.TypeAlias = tuple[ta.Callable[[], str]]

LinecacheEntry: ta.TypeAlias = LinecacheKey | LinecacheLazyLoader


class LinecacheProtocol(ta.Protocol):
    @property
    def cache(self) -> ta.MutableMapping[str, LinecacheEntry]: ...

    def clearcache(self) -> None: ...

    def getline(
        self,
        filename: str,
        lineno: int,
        module_globals: ta.Mapping[str, ta.Any] | None = None,
    ) -> str: ...

    def getlines(
        self,
        filename: str,
        module_globals: ta.Mapping[str, ta.Any] | None = None,
    ) -> ta.Sequence[str]: ...

    def checkcache(
        self,
        filename: str | None = None,
    ) -> None: ...

    def updatecache(
        self,
        filename: str,
        module_globals: ta.Mapping[str, ta.Any] | None = None,
    ) -> ta.Sequence[str]: ...

    def lazycache(
        self,
        filename: str,
        module_globals: ta.Mapping[str, ta.Any],
    ) -> bool: ...



import typing as ta

from .. import dataclasses as dc
from .. import lang


T = ta.TypeVar('T')


##


class TypedValue(lang.Abstract):
    def __bool__(self) -> ta.NoReturn:
        raise TypeError(f"Cannot convert {self!r} to bool - use '.v' or 'is not None'.")


##


_UNIQUE_BASES: set[type[TypedValue]] = set()


class UniqueTypedValue(TypedValue, lang.Abstract):
    _unique_typed_value_cls: ta.ClassVar[type[TypedValue]]

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        if any(ur in cls.__bases__ for ur in _UNIQUE_BASES):
            try:
                cls._unique_typed_value_cls  # noqa
            except AttributeError:
                cls._unique_typed_value_cls = cls
            else:
                raise TypeError(f'Class already has _unique_typed_value_cls: {cls}')


##


class ScalarTypedValue(TypedValue, dc.Box[T], abstract=True):
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        if UniqueTypedValue in (mro := cls.__mro__) and mro.index(ScalarTypedValue) > mro.index(UniqueTypedValue):
            raise TypeError(f'Class {cls} must not have UniqueTypedValue before ScalarTypedValue in mro')


##


class UniqueScalarTypedValue(ScalarTypedValue[T], UniqueTypedValue, abstract=True):
    pass


##


_UNIQUE_BASES.update([
    UniqueTypedValue,
    UniqueScalarTypedValue,
])



"""
TODO:
 - BufferedBytesCoroReader
 - docstrings
 - memoryviews
"""
import abc
import typing as ta

from ... import check
from ... import lang
from .consts import DEFAULT_BUFFER_SIZE


T = ta.TypeVar('T')

O = ta.TypeVar('O')
I = ta.TypeVar('I')
R = ta.TypeVar('R')

AnyT = ta.TypeVar('AnyT', bound=ta.Any)


# Reader coros yield ints of amounts of data needed, or None for needing any amount of data.
ReaderCoro: ta.TypeAlias = ta.Generator[int | None, I, R]

BytesReaderCoro: ta.TypeAlias = ReaderCoro[bytes, R]
StrReaderCoro: ta.TypeAlias = ReaderCoro[str, R]


# Exact reader coros always specify read sizes.
ExactReaderCoro: ta.TypeAlias = ta.Generator[int, I, R]

BytesExactReaderCoro: ta.TypeAlias = ExactReaderCoro[bytes, R]
StrExactReaderCoro: ta.TypeAlias = ExactReaderCoro[str, R]


##


class _BytesJoiner:
    def _join(self, lst: list[bytes]) -> bytes:
        return b''.join(lst)


class _StrJoiner:
    def _join(self, lst: list[str]) -> str:
        return ''.join(lst)


##


class CoroReader(lang.Abstract, ta.Generic[T]):
    @abc.abstractmethod
    def read(self, sz: int | None) -> ReaderCoro[T, T]:
        raise NotImplementedError

    def read_exact(self, sz: int) -> ReaderCoro[T, T]:
        d: ta.Any = yield from self.read(sz)
        if len(d) != sz:
            raise EOFError(f'CoroReader got {len(d)}, expected {sz}')
        return d


##


class PrependableCoroReader(CoroReader[AnyT], lang.Abstract):
    def __init__(self) -> None:
        super().__init__()

        self._queue: list[tuple[AnyT, int]] = []

    @abc.abstractmethod
    def _join(self, lst: list[AnyT]) -> AnyT:
        raise NotImplementedError

    def read(self, sz: int | None) -> ReaderCoro[AnyT, AnyT]:
        if not self._queue:
            d: AnyT = check.not_none((yield sz))
            return d

        if sz is None:
            return self._queue.pop(0)[0]

        lst: list[AnyT] = []
        rem = sz
        while rem > 0 and self._queue:
            c, p = self._queue[0]

            if len(c) - p > rem:
                lst.append(c[p:p + rem])
                self._queue[0] = (c, p + rem)
                return self._join(lst)

            lst.append(c[p:])
            rem -= len(c) - p
            self._queue.pop(0)

        if rem:
            d = check.not_none((yield rem))
            if d:
                lst.append(d)

        if len(lst) == 1:
            return lst[0]
        else:
            return self._join(lst)

    def prepend(self, d: AnyT, p: int | None = None) -> None:
        if d:
            self._queue.insert(0, (d, p or 0))


class PrependableBytesCoroReader(
    _BytesJoiner,
    PrependableCoroReader[bytes],
):
    pass


class PrependableStrCoroReader(
    _StrJoiner,
    PrependableCoroReader[str],
):
    pass


prependable_bytes_coro_reader = PrependableBytesCoroReader
prependable_str_coro_reader = PrependableStrCoroReader


##


class BufferedCoroReader(PrependableCoroReader[AnyT], lang.Abstract):
    def __init__(
            self,
            buffer_size: int = DEFAULT_BUFFER_SIZE,
    ) -> None:
        check.arg(buffer_size > 0)

        super().__init__()

        self._buffer_size = buffer_size

    def read(self, sz: int | None) -> ReaderCoro[AnyT, AnyT]:
        g = super().read(sz)
        i: ta.Any = None
        while True:
            try:
                q = g.send(i)
            except StopIteration as e:
                return e.value

            check.state(not self._queue)

            if q is None:
                i = check.not_none((yield None))
                continue

            r = max(q, self._buffer_size)
            d: AnyT = check.not_none((yield r))
            if len(d) < q:
                i = d
                continue

            i = d[:q]
            self.prepend(d, q)


class BufferedBytesCoroReader(
    _BytesJoiner,
    BufferedCoroReader[bytes],
    PrependableCoroReader[bytes],
):
    pass


class BufferedStrCoroReader(
    _StrJoiner,
    BufferedCoroReader[str],
    PrependableCoroReader[str],
):
    pass


buffered_bytes_coro_reader = BufferedBytesCoroReader
buffered_str_coro_reader = BufferedStrCoroReader



import dataclasses as dc
import typing as ta

from .... import check
from ..._internals import STD_FIELDS_ATTR
from ...debug import DEBUG
from ...inspect import FieldsInspection
from ...specs import FieldSpec
from ...specs import FieldType
from ..generation.idents import IDENT_PREFIX
from ..processing.base import ProcessingContext
from ..processing.base import Processor
from ..processing.priority import ProcessorPriority
from ..processing.registry import register_processing_context_item_factory
from ..processing.registry import register_processor_type


##


class InstanceFields(list[FieldSpec]):
    pass


def get_instance_fields(fields: ta.Iterable[FieldSpec]) -> InstanceFields:
    return InstanceFields([f for f in fields if f.field_type is FieldType.INSTANCE])


@register_processing_context_item_factory(InstanceFields)
def _instance_fields_processing_context_item_factory(ctx: ProcessingContext) -> InstanceFields:
    return get_instance_fields(ctx.cs.fields)


##


class InitFields(ta.NamedTuple):
    all: ta.Sequence[FieldSpec]
    ordered: ta.Sequence[FieldSpec]
    std: ta.Sequence[FieldSpec]
    kw_only: ta.Sequence[FieldSpec]


def calc_init_fields(
        fields: ta.Iterable[FieldSpec],
        *,
        reorder: bool,
        class_kw_only: bool,
) -> InitFields:
    all_init_fields = [
        f
        for f in fields
        if f.field_type in (FieldType.INSTANCE, FieldType.INIT_VAR)
    ]

    def f_kw_only(f: FieldSpec) -> bool:
        return f.kw_only if f.kw_only is not None else class_kw_only

    ordered_init_fields = list(all_init_fields)
    if reorder:
        ordered_init_fields.sort(key=lambda f: (f.default.present, not f_kw_only(f)))

    std_init_fields = tuple(f1 for f1 in ordered_init_fields if f1.init and not f_kw_only(f1))
    kw_only_init_fields = tuple(f1 for f1 in ordered_init_fields if f1.init and f_kw_only(f1))

    return InitFields(
        all=all_init_fields,
        ordered=ordered_init_fields,
        std=std_init_fields,
        kw_only=kw_only_init_fields,
    )


@register_processing_context_item_factory(InitFields)
def _init_fields_processing_context_item_factory(ctx: ProcessingContext) -> InitFields:
    return calc_init_fields(
        ctx.cs.fields,
        reorder=ctx.cs.reorder,
        class_kw_only=ctx.cs.kw_only,
    )


##


StdFields = ta.NewType('StdFields', ta.Mapping[str, dc.Field])


@register_processing_context_item_factory(StdFields)
def _std_fields_processing_context_item_factory(ctx: ProcessingContext) -> StdFields:
    fld_dct = ctx.cls.__dict__[STD_FIELDS_ATTR]
    for fn, f in fld_dct.items():
        if DEBUG:
            check.isinstance(f, dc.Field)
        check.equal(f.name, fn)
    check.equal(set(fld_dct), set(ctx.cs.fields_by_name))
    return StdFields(fld_dct)


##


@register_processing_context_item_factory(FieldsInspection)
def _fields_inspection_processing_context_item_factory(ctx: ProcessingContext) -> FieldsInspection:
    return FieldsInspection(
        ctx.cls,
        cls_fields=ctx[StdFields],
    )


##


@register_processor_type(priority=ProcessorPriority.BOOTSTRAP)
class FieldsProcessor(Processor):
    def check(self) -> None:
        check.not_none(self._ctx[StdFields])
        for f in self._ctx.cs.fields:
            check.arg(not f.name.startswith(IDENT_PREFIX))



import dataclasses as dc
import enum
import typing as ta

from ... import check
from ... import reflect as rfl
from ..base.contexts import MarshalContext
from ..base.contexts import MarshalFactoryContext
from ..base.contexts import UnmarshalContext
from ..base.contexts import UnmarshalFactoryContext
from ..base.types import Marshaler
from ..base.types import MarshalerFactory
from ..base.types import Unmarshaler
from ..base.types import UnmarshalerFactory
from ..base.values import Value


##


@dc.dataclass(frozen=True)
class EnumMarshaler(Marshaler):
    ty: type[enum.Enum]

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return o.name


class EnumMarshalerFactory(MarshalerFactory):
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not (isinstance(rty, type) and issubclass(rty, enum.Enum)):
            return None
        return lambda: EnumMarshaler(rty)  # noqa


@dc.dataclass(frozen=True)
class EnumUnmarshaler(Unmarshaler):
    ty: type[enum.Enum]

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return self.ty[check.isinstance(v, str)]


class EnumUnmarshalerFactory(UnmarshalerFactory):
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not (isinstance(rty, type) and issubclass(rty, enum.Enum)):
            return None
        return lambda: EnumUnmarshaler(rty)  # noqa



import dataclasses as dc
import typing as ta

from .... import lang
from .globals import FnGlobal
from .idents import IDENT_PREFIX


T = ta.TypeVar('T')


##


class _OpRef(ta.NamedTuple, ta.Generic[T]):
    name: str


@ta.final
class OpRef(_OpRef[T]):
    def __repr__(self) -> str:
        return f'OpRef(name={self.name!r})'

    #

    _ident: str

    def ident(self) -> str:
        try:
            return self._ident
        except AttributeError:
            pass

        ident = IDENT_PREFIX + self.name.replace('.', '__')
        self._ident = ident  # noqa
        return ident


##


OpRefMap: ta.TypeAlias = ta.Mapping[OpRef, ta.Any]

Ref: ta.TypeAlias = OpRef | FnGlobal


##


IfAttrPresent: ta.TypeAlias = ta.Literal['skip', 'replace', 'error']


@dc.dataclass(frozen=True)
class Op(lang.Abstract):
    pass


@dc.dataclass(frozen=True)
class SetAttrOp(Op):
    name: str
    value: Ref | ta.Any

    if_present: IfAttrPresent = dc.field(default='replace', kw_only=True)


@dc.dataclass(frozen=True)
class AddMethodOp(Op):
    name: str
    src: str
    refs: frozenset[Ref] = dc.field(default=frozenset())

    if_present: IfAttrPresent = dc.field(default='error', kw_only=True)


@dc.dataclass(frozen=True)
class AddPropertyOp(Op):
    name: str
    get_src: str | None = None
    set_src: str | None = None
    del_src: str | None = None
    refs: frozenset[Ref] = frozenset()


##


def get_op_refs(op: Op) -> frozenset[Ref]:
    if isinstance(op, SetAttrOp):
        if isinstance(v := op.value, (OpRef, FnGlobal)):
            return frozenset([v])
        else:
            return frozenset()

    elif isinstance(op, (AddMethodOp, AddPropertyOp)):
        return op.refs

    else:
        raise TypeError(op)



from .. import check
from .. import dataclasses as dc
from .. import lang
from .elements import Element
from .keys import Key
from .providers import Provider
from .types import Scope
from .types import Unscoped


##


@dc.dataclass(frozen=True)
@dc.extra_class_params(cache_hash=True)
class Binding(Element, lang.Final):
    key: Key = dc.xfield(coerce=check.of_isinstance(Key))
    provider: Provider = dc.xfield(coerce=check.of_isinstance(Provider))
    scope: Scope = dc.xfield(default=Unscoped(), coerce=check.of_isinstance(Scope))



import typing as ta

from ... import reflect as rfl
from ..base.contexts import MarshalContext
from ..base.contexts import UnmarshalContext
from ..base.types import Marshaler
from ..base.types import Unmarshaler
from ..base.values import Value
from ..factories.typemap import TypeMapMarshalerFactory
from ..factories.typemap import TypeMapUnmarshalerFactory


##


class AnyMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return ctx.marshal_factory_context.make_marshaler(type(o)).marshal(ctx, o)

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return v


ANY_MARSHALER_UNMARSHALER = AnyMarshalerUnmarshaler()

ANY_MARSHALER_FACTORY = TypeMapMarshalerFactory({rfl.ANY: ANY_MARSHALER_UNMARSHALER})
ANY_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({rfl.ANY: ANY_MARSHALER_UNMARSHALER})



from .abc import (  # noqa
    DbapiTypeCode,
    DbapiColumnDescription,
    DbapiColumnDescription_,
    DbapiConnection,
    DbapiCursor,
    DbapiThreadSafety,
    DbapiModule,
)

from .drivers import (  # noqa
    DbapiDialect,
    DbapiDriver,
    DbapiDrivers,
)



import dataclasses as dc
import typing as ta

from ...specs import FieldType
from ..generation.base import Generator
from ..generation.base import Plan
from ..generation.base import PlanResult
from ..generation.idents import CLS_IDENT
from ..generation.ops import AddMethodOp
from ..generation.ops import Op
from ..generation.registry import register_generator_type
from ..generation.utils import build_attr_kwargs_body_src_lines
from ..processing.base import ProcessingContext


##


@dc.dataclass(frozen=True)
class CopyPlan(Plan):
    fields: tuple[str, ...]


@register_generator_type(CopyPlan)
class CopyGenerator(Generator[CopyPlan]):
    def plan(self, ctx: ProcessingContext) -> PlanResult[CopyPlan] | None:
        if '__copy__' in ctx.cls.__dict__:
            return None

        return PlanResult(CopyPlan(
            tuple(f.name for f in ctx.cs.fields if f.field_type is not FieldType.CLASS_VAR),
        ))

    def generate(self, pl: CopyPlan) -> ta.Iterable[Op]:
        return_lines: list[str]
        if pl.fields:
            return_lines = [
                f'    return {CLS_IDENT}(  # noqa',
                *build_attr_kwargs_body_src_lines(
                    'self',
                    *pl.fields,
                    prefix='        ',
                ),
                f'    )',
            ]
        else:
            return_lines = [
                f'    return {CLS_IDENT}()  # noqa',
            ]

        lines = [
            f'def __copy__(self):',
            f'    if self.__class__ is not {CLS_IDENT}:',
            f'        raise TypeError(self)',
            *return_lines,
        ]

        return [
            AddMethodOp(
                '__copy__',
                '\n'.join(lines),
            ),
        ]



"""
A small but likely growing collection of (completely optional) tools to make pydevd (PyCharm's, among other python
IDE's, debugger) do hard things. Originally explored and added to get spark jvm python subprocesses to connect back to
an already-debugging PyCharm instance to debug PySpark jobs.

TODO:
 - https://www.jetbrains.com/help/pycharm/remote-debugging-with-product.html#

====

https://www.jetbrains.com/help/pycharm/remote-debugging-with-product.html#remote-debug-config ->

pycharm_port = 43251
pycharm_version = '241.18034.82'
buf = textwrap.dedent(f'''
    import subprocess
    import sys
    subprocess.check_call([sys.executable, '-mpip', 'install', f'pydevd-pycharm~={pycharm_version}'])

    import pydevd_pycharm  # noqa
    pydevd_pycharm.settrace(
        'docker.for.mac.localhost',
         port={pycharm_port},
          stdoutToServer=True,
           stderrToServer=True,
       )
''') + '\n' * 2 + buf

====

TODO: monkeypatch:

/Applications/PyCharm.app/Contents/plugins/python-ce/helpers/pydev/_pydev_bundle/pydev_monkey.py ::

def starts_with_python_shebang(path):
    try:
        with open(path) as f:
            for line in f:
                line = line.strip()
                if line:
                    for name in PYTHON_NAMES:
                        if line.startswith('#!/usr/bin/env %s' % name):
                            return True
                    return False
    except (UnicodeDecodeError, IsADirectoryError):  # <-- Add catch for `IsADirectoryError`
        return False
    except:
        traceback.print_exc()
        return False
"""
import json
import os
import sys
import tempfile
import textwrap
import types
import typing as ta

from .. import check
from .. import lang


##


ALLOW_DEBUGGER_CALLS = False


DEBUGGER_CALL_PACKAGES = {
    '_pydevd_bundle',
}


def is_debugger_call(hoist: int = 0, walk: int = 2) -> bool:
    frame: ta.Optional[types.FrameType] = sys._getframe(2 + hoist)  # noqa
    for _ in range(walk):
        if frame is None:
            break
        package = frame.f_globals.get('__package__')
        if package in DEBUGGER_CALL_PACKAGES:
            return True
        frame = frame.f_back
    return False


class DebuggerCallForbiddenError(Exception):
    pass


def forbid_debugger_call(hoist: int = 0) -> None:
    # FIXME: only reentrant?
    if not ALLOW_DEBUGGER_CALLS and is_debugger_call(hoist + 1):
        raise DebuggerCallForbiddenError


##


@lang.cached_function
def silence_subprocess_check() -> None:
    try:
        # /Applications/PyCharm.app/Contents/plugins/python/helpers/pydev/_pydev_bundle/pydev_monkey.py
        from _pydev_bundle import pydev_monkey  # noqa
    except ImportError:
        return

    new_tb = lang.proxy_import('traceback')
    new_tb.print_exc = lambda *a, **k: None  # type: ignore  # noqa
    pydev_monkey.traceback = new_tb


@lang.cached_function
def patch_for_trio_asyncio() -> None:
    """Fix for `trio a callable object was expected by call_soon(), got Task`"""

    try:
        import pydevd_nest_asyncio  # noqa
    except ImportError:
        return

    import trio_asyncio._base  # noqa

    def new_call_soon(self, callback, *args, **context):
        _, callback = pydevd_nest_asyncio._PydevdAsyncioUtils.try_to_get_internal_callback(callback)  # noqa
        return orig_call_soon(self, callback, *args, **context)

    orig_call_soon = trio_asyncio._base.BaseTrioEventLoop.call_soon  # noqa
    trio_asyncio._base.BaseTrioEventLoop.call_soon = new_call_soon  # noqa


##


@lang.cached_function
def _pydevd() -> types.ModuleType | None:
    try:
        return __import__('pydevd')
    except ImportError:
        return None


def is_present() -> bool:
    # FIXME: try to use `lang.can_import('pydevd'), but raises with:
    # INTERNALERROR>   File "/Users/spinlock/src/wrmsr/omlish/omlish/lang/imports/resolving.py", line 16, in can_import
    # INTERNALERROR>     spec = importlib.util.find_spec(name, package)
    # INTERNALERROR>   File "<frozen importlib.util>", line 111, in find_spec
    # INTERNALERROR> ValueError: pydevd.__spec__ is None
    # Really want to avoid actually importing pydevd due to side-effects, slowness, and even a pkg_resources deprecation
    # warning...
    try:
        __import__('pydevd')
    except ImportError:
        return False
    else:
        return True


def get_setup() -> dict | None:
    if is_present():
        return check.not_none(_pydevd()).SetupHolder.setup
    else:
        return None


def is_running() -> bool:
    return get_setup() is not None


#


@lang.cached_function
def _pydevd_constants() -> types.ModuleType | None:
    try:
        return __import__('_pydevd_bundle.pydevd_constants').pydevd_constants
    except ImportError:
        return None


def get_global_debugger() -> ta.Any | None:
    if (pc := _pydevd_constants()) is None:
        return None

    return pc.get_global_debugger()


##


ARGS_ENV_VAR = 'PYDEVD_ARGS'


def get_args() -> list[str]:
    check.state(is_present())
    setup: ta.Mapping[ta.Any, ta.Any] = check.isinstance(get_setup(), dict)
    args = [check.not_none(check.not_none(_pydevd()).__file__)]

    for k in [
        'port',
        'vm_type',
        'client',
    ]:
        if v := setup[k]:
            args.extend(['--' + k, str(v)])

    for k in [
        'server',
        'multiproc',
        'multiprocess',
        'save-signatures',
        'save-threading',
        'save-asyncio',
        'print-in-debugger-startup',
        'cmd-line',
    ]:
        if setup[k]:
            args.append('--' + k)

    if setup['qt-support']:
        args.append('--qt-support=' + setup['qt-support'])

    return args


def save_args() -> None:
    if is_present():
        os.environ[ARGS_ENV_VAR] = json.dumps(get_args())


def maybe_reexec(
        *,
        file: str | None = None,
        module: str | None = None,
        silence: bool = False,
) -> None:
    if ARGS_ENV_VAR not in os.environ:
        return

    try:
        import pydevd  # noqa
    except ImportError:
        return

    if pydevd.SetupHolder.setup is not None:  # noqa
        return

    if module is not None:
        if file is not None:
            raise ValueError

        tmpdir = tempfile.mkdtemp()
        bootstrap_path = os.path.join(tmpdir, 'bootstrap.py')
        with open(bootstrap_path, 'w') as f:
            f.write(textwrap.dedent(f"""
            import sys
            old_paths = set(sys.path)
            for new_path in {sys.path!r}:
                if new_path not in old_paths:
                    sys.path.insert(0, new_path)

            import runpy
            runpy.run_module({module!r}, run_name='__main__')
            """))
        file = bootstrap_path

    elif file is None:
        raise ValueError

    args = [sys.executable]
    args.extend(json.loads(os.environ[ARGS_ENV_VAR]))
    args.extend(['--file', file])
    args.extend(sys.argv[1:])

    if silence:
        tmpdir = tempfile.mkdtemp()
        bootstrap_path = os.path.join(tmpdir, 'bootstrap.py')
        with open(bootstrap_path, 'w') as f:
            f.write(textwrap.dedent(f"""
                import sys
                old_paths = set(sys.path)
                for new_path in {sys.path!r}:
                    if new_path not in old_paths:
                        sys.path.insert(0, new_path)

                _stderr_write = sys.stderr.write
                def stderr_write(*args, **kwargs):
                    code = sys._getframe(1).f_code
                    if code is not None and code.co_filename and code.co_filename.endswith('/pydev_log.py'):
                        return
                    _stderr_write(*args, **kwargs)
                sys.stderr.write = stderr_write

                sys.argv = {args[1:]!r}
                import runpy
                runpy.run_path({args[1]!r}, run_name='__main__')
            """))
        args = [args[0], bootstrap_path]

    os.execvp(sys.executable, args)


def debug_unhandled_exception(exc_info: ta.Any = None) -> None:
    if exc_info is None:
        exc_info = sys.exc_info()

    try:
        import pydevd

    except ImportError:
        return

    et, e, tb = exc_info

    while tb.tb_next is not None:
        tb = tb.tb_next
    original_frame = tb.tb_frame

    pydevd.settrace(stop_at_frame=original_frame, suspend=True)



import typing as ta


##


class ProtocolForbiddenAsBaseClass(ta.Protocol):
    pass


class ProtocolForbiddenAsBaseClassTypeError(TypeError):
    pass


def _ProtocolForbiddenAsBaseClass__init_subclass__(cls: ta.Any, **kwargs: ta.Any) -> None:  # noqa
    if ta.Protocol not in cls.__mro__:
        raise TypeError(f'Class {cls} must be a Protocol')

    super(ProtocolForbiddenAsBaseClass, cls).__init_subclass__(**kwargs)  # noqa

    # TODO: ta.Protocol not in cls.__bases__ ?
    if not cls.__dict__['_is_protocol']:
        raise ProtocolForbiddenAsBaseClassTypeError(cls)


setattr(ProtocolForbiddenAsBaseClass, '__init_subclass__', classmethod(_ProtocolForbiddenAsBaseClass__init_subclass__))



"""
TODO:
 - clamp down on as_ident / CanIdent - no strs allowed
"""
import abc
import functools
import typing as ta

from ... import cached
from ... import lang
from ..qualifiedname import QualifiedName
from .base import Builder
from .base import HasQn
from .base import Node


##


class IdentLike(abc.ABC):  # noqa
    pass


##


class Ident(Node, IdentLike, HasQn, lang.Final):
    s: str

    @cached.property
    def qn(self) -> QualifiedName:
        return QualifiedName((self.s,))


##


@functools.singledispatch
def as_ident(o: ta.Any) -> Ident:
    raise TypeError(o)


@as_ident.register
def _(i: Ident) -> Ident:
    return i


@as_ident.register
def _(s: str) -> Ident:
    return Ident(s)


##


CanIdent: ta.TypeAlias = IdentLike | Ident | str


class IdentAccessor(lang.Final):
    def __getattr__(self, s: str) -> Ident:
        return Ident(s)

    def __call__(self, o: CanIdent) -> Ident:
        return as_ident(o)


##


class IdentBuilder(Builder):
    @ta.final
    def ident(self, o: CanIdent) -> Ident:
        return as_ident(o)

    @ta.final
    @property
    def i(self) -> IdentAccessor:
        return IdentAccessor()



import abc
import contextlib
import typing as ta
import weakref

from .impls import find_impl as default_find_impl


T = ta.TypeVar('T')


##


class Dispatcher(ta.Generic[T]):
    """Shared dispatching system for functions and methods. Logic directly mimics `functools.singledispatch`."""

    def __init__(self, find_impl: ta.Callable[[type, ta.Mapping[type, T]], T | None] | None = None) -> None:
        super().__init__()

        if find_impl is None:
            find_impl = default_find_impl

        impls_by_arg_cls: dict[type, T] = {}
        self._impls_by_arg_cls = impls_by_arg_cls

        dispatch_cache: dict[ta.Any, T | None] = {}
        self._get_dispatch_cache = lambda: dispatch_cache

        def cache_remove(k, self_ref=weakref.ref(self)):
            if (ref_self := self_ref()) is not None:
                cache = ref_self._get_dispatch_cache()  # noqa
                with contextlib.suppress(KeyError):
                    del cache[k]

        cache_token: ta.Any = None
        self._get_cache_token = lambda: cache_token

        weakref_ref_ = weakref.ref

        def dispatch(cls: type) -> T | None:
            nonlocal cache_token

            if cache_token is not None and (current_token := abc.get_cache_token()) != cache_token:
                dispatch_cache.clear()
                cache_token = current_token

            cls_ref = weakref_ref_(cls)
            try:
                return dispatch_cache[cls_ref]
            except KeyError:
                pass
            del cls_ref

            impl: T | None
            try:
                impl = impls_by_arg_cls[cls]
            except KeyError:
                impl = find_impl(cls, impls_by_arg_cls)

            dispatch_cache[weakref_ref_(cls, cache_remove)] = impl
            return impl

        self.dispatch = dispatch

        def register(impl: T, cls_col: ta.Iterable[type]) -> T:
            nonlocal cache_token

            for cls in cls_col:
                impls_by_arg_cls[cls] = impl

                if cache_token is None and hasattr(cls, '__abstractmethods__'):
                    cache_token = abc.get_cache_token()

            dispatch_cache.clear()
            return impl

        self.register = register

    _get_cache_token: ta.Callable[[], int]
    _get_dispatch_cache: ta.Callable[[], ta.Any]

    def cache_size(self) -> int:
        return len(self._get_dispatch_cache())

    dispatch: ta.Callable[[type], T | None]

    register: ta.Callable[[T, ta.Iterable[type]], T]


##


# from x.dispatch import _gpto1_2 as _dispatch  # noqa
#
#
# class Dispatcher(ta.Generic[T]):  # noqa
#     def __init__(self, find_impl: ta.Callable[[type, ta.Mapping[type, T]], T | None] | None = None) -> None:
#         super().__init__()
#
#         if find_impl is None:
#             find_impl = default_find_impl
#         self._x = _dispatch.Dispatcher(find_impl)
#
#     def cache_size(self) -> int:
#         return self._x.cache_size()
#
#     def dispatch(self, cls: type) -> T | None:
#         return self._x.dispatch(cls)
#
#     def register(self, impl: T, cls_col: ta.Iterable[type]) -> T:
#         return self._x.register(impl, cls_col)



import typing as ta

import anyio

from .utils import get_current_task


##


BackendTask: ta.TypeAlias = ta.Union[  # noqa
    # asyncio.tasks.Task,
    # trio.lowlevel.Task,
    ta.Any,
]


def _is_class_named(obj: ta.Any, m: str, n: str) -> bool:
    cls = obj.__class__
    return cls.__module__ == m and cls.__name__ == n


def get_backend_task(at: anyio.TaskInfo) -> BackendTask | None:
    if _is_class_named(at, 'anyio._backends._asyncio', 'AsyncIOTaskInfo'):
        # https://github.com/agronholm/anyio/blob/8907964926a24461840eee0925d3f355e729f15d/src/anyio/_backends/_asyncio.py#L1846  # noqa
        # weakref.ref
        obj = at._task()  # type: ignore  # noqa
        if obj is not None and not (
                _is_class_named(obj, '_asyncio', 'Task') or
                _is_class_named(obj, 'asyncio.tasks', 'Task')
        ):
            raise TypeError(obj)
        return obj

    elif _is_class_named(at, 'anyio._backends._trio', 'TrioTaskInfo'):
        # https://github.com/agronholm/anyio/blob/8907964926a24461840eee0925d3f355e729f15d/src/anyio/_backends/_trio.py#L850  # noqa
        # weakref.proxy
        # https://stackoverflow.com/a/62144308 :|
        obj = at._task.__repr__.__self__  # type: ignore  # noqa
        if obj is not None and not _is_class_named(obj, 'trio.lowlevel', 'Task'):
            raise TypeError(obj)
        return obj

    else:
        raise TypeError(at)


def get_current_backend_task() -> BackendTask | None:
    if (at := get_current_task()) is not None:
        return get_backend_task(at)
    else:
        return None



# ruff: noqa: UP006 UP036 UP045
import functools
import sys
import types
import typing as ta


T = ta.TypeVar('T')


##


_GENERIC_ALIAS_TYPES = (
    ta._GenericAlias,  # type: ignore  # noqa
    *([ta._SpecialGenericAlias] if hasattr(ta, '_SpecialGenericAlias') else []),  # noqa
)


def is_generic_alias(obj: ta.Any, *, origin: ta.Any = None) -> bool:
    return (
        isinstance(obj, _GENERIC_ALIAS_TYPES) and
        (origin is None or ta.get_origin(obj) is origin)
    )


is_callable_alias = functools.partial(is_generic_alias, origin=ta.Callable)


##


_UNION_ALIAS_ORIGINS = frozenset([
    ta.get_origin(ta.Optional[int]),
    *(
        [
            ta.get_origin(int | None),
            ta.get_origin(getattr(ta, 'TypeVar')('_T') | None),
        ] if sys.version_info >= (3, 10) else ()
    ),
])


def is_union_alias(obj: ta.Any) -> bool:
    return ta.get_origin(obj) in _UNION_ALIAS_ORIGINS


#


def is_optional_alias(spec: ta.Any) -> bool:
    return (
        is_union_alias(spec) and
        len(ta.get_args(spec)) == 2 and
        any(a in (None, type(None)) for a in ta.get_args(spec))
    )


def get_optional_alias_arg(spec: ta.Any) -> ta.Any:
    [it] = [it for it in ta.get_args(spec) if it not in (None, type(None))]
    return it


##


def is_new_type(spec: ta.Any) -> bool:
    if isinstance(ta.NewType, type):
        return isinstance(spec, ta.NewType)
    else:
        # Before https://github.com/python/cpython/commit/c2f33dfc83ab270412bf243fb21f724037effa1a
        return isinstance(spec, types.FunctionType) and spec.__code__ is ta.NewType.__code__.co_consts[1]  # type: ignore  # noqa


def get_new_type_supertype(spec: ta.Any) -> ta.Any:
    return spec.__supertype__


##


def is_literal_type(spec: ta.Any) -> bool:
    if hasattr(ta, '_LiteralGenericAlias'):
        return isinstance(spec, ta._LiteralGenericAlias)  # noqa
    else:
        return (
            isinstance(spec, ta._GenericAlias) and  # type: ignore  # noqa
            spec.__origin__ is ta.Literal
        )


def get_literal_type_args(spec: ta.Any) -> ta.Iterable[ta.Any]:
    return spec.__args__



"""
FIXME:
 - lol shutdown deadlocks
 - whole thing is just gross is this its own thread or what
 - cli with client helper too

TODO:
 - !!! ANYIO !!!
 - optional paramiko ssh-server
 - optional ipython embed
 - https://github.com/python/cpython/tree/56470004e58911b146c016fc9fec4461b8f69454/Lib/_pyrepl

See:
 - https://github.com/vxgmichel/aioconsole/blob/e55f4b0601da3b3a40a88c965526d35ab38b5841/aioconsole/server.py
 - https://github.com/nhoad/aiomanhole
 - https://github.com/twisted/twisted/blob/00aa56f5257060304d41f09651c6ab58ee6104d6/src/twisted/conch/manhole.py
  - https://github.com/Yelp/Tron/blob/4b864a73bd129b03e9890c134212972452bc6ab0/tron/manhole.py#L8
 - https://github.com/ionelmc/python-manhole
 - https://github.com/python/cpython/tree/15d48aea02099ffc5bdc5511cc53ced460cb31b9/Lib/_pyrepl

socat - UNIX-CONNECT:repl.sock
"""
import contextlib
import functools
import os
import socket as sock
import threading
import typing as ta
import weakref

from ... import check
from ... import dataclasses as dc
from ...logs import all as logs
from .console import InteractiveSocketConsole


log = logs.get_module_logger(globals())


##


class ReplServer:
    CONNECTION_THREAD_NAME = 'ReplServerConnection'

    @dc.dataclass(frozen=True)
    class Config:
        path: str
        file_mode: int | None = 0o660
        poll_interval: float = 0.5
        exit_timeout: float = 10.0

    def __init__(
            self,
            config: Config,
    ) -> None:
        super().__init__()

        check.not_empty(config.path)
        self._config = check.isinstance(config, ReplServer.Config)

        self._socket: sock.socket | None = None
        self._is_running = False
        self._consoles_by_threads: ta.MutableMapping[threading.Thread, InteractiveSocketConsole] = \
            weakref.WeakKeyDictionary()  # noqa
        self._is_shutdown = threading.Event()
        self._should_shutdown = False

    @property
    def path(self) -> str:
        return self._config.path

    def __enter__(self) -> ta.Self:
        check.state(not self._is_running)
        check.state(not self._is_shutdown.is_set())
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        if not self._is_shutdown.is_set():
            self.shutdown(True, self._config.exit_timeout)

    def run(self) -> None:
        check.state(not self._is_running)
        check.state(not self._is_shutdown.is_set())

        if os.path.exists(self._config.path):
            os.unlink(self._config.path)

        self._socket = sock.socket(sock.AF_UNIX, sock.SOCK_STREAM)
        self._socket.settimeout(self._config.poll_interval)

        if self._config.file_mode is not None:
            prev_umask = os.umask(~self._config.file_mode)
        else:
            prev_umask = None
        try:
            self._socket.bind(self._config.path)
        finally:
            if prev_umask is not None:
                os.umask(prev_umask)

        with contextlib.closing(self._socket):
            self._socket.listen(1)

            log.info('Repl server listening on file %s', self._config.path)

            self._is_running = True
            try:
                while not self._should_shutdown:
                    try:
                        conn, _ = self._socket.accept()
                    except TimeoutError:
                        continue

                    log.info('Got repl server connection on file %s', self._config.path)

                    def run(conn):
                        with contextlib.closing(conn):
                            variables = globals().copy()

                            console = InteractiveSocketConsole(conn, variables)
                            variables['__console__'] = console

                            log.info(
                                'Starting console %x repl server connection on file %s on thread %r',
                                id(console),
                                self._config.path,
                                threading.current_thread().ident,
                            )
                            self._consoles_by_threads[threading.current_thread()] = console
                            console.interact()

                    thread = threading.Thread(
                        target=functools.partial(run, conn),
                        daemon=True,
                        name=self.CONNECTION_THREAD_NAME)
                    thread.start()

                for console in self._consoles_by_threads.values():
                    try:
                        console.conn.close()
                    except Exception:
                        log.exception('Error shutting down')

                for thread in self._consoles_by_threads:
                    try:
                        thread.join(self._config.exit_timeout)
                    except Exception:
                        log.exception('Error shutting down')

                os.unlink(self._config.path)

            finally:
                self._is_shutdown.set()
                self._is_running = False

    def shutdown(self, block: bool = False, timeout: float | None = None) -> None:
        self._should_shutdown = True
        if block:
            self._is_shutdown.wait(timeout=timeout)


def run() -> None:
    with ReplServer(ReplServer.Config('repl.sock')) as repl_server:
        repl_server.run()



# ruff: noqa: UP006 UP045
# @omlish-lite
"""
https://docs.python.org/3/howto/logging.html#configuring-logging
https://docs.python.org/3/library/logging.config.html#logging-config-dictschema
"""
import dataclasses as dc
import typing as ta


FilterDictLoggingConfig = ta.Dict[str, ta.Any]  # ta.TypeAlias
FormatterDictLoggingConfig = ta.Dict[str, ta.Any]  # ta.TypeAlias
HandlerDictLoggingConfig = ta.Dict[str, ta.Any]  # ta.TypeAlias
LoggerDictLoggingConfig = ta.Dict[str, ta.Any]  # ta.TypeAlias


##


@dc.dataclass()
class DictLoggingConfig:
    version: int = 1
    incremental: bool = False
    disable_existing_loggers: bool = False
    filters: ta.Dict[str, FilterDictLoggingConfig] = dc.field(default_factory=dict)
    formatters: ta.Dict[str, FormatterDictLoggingConfig] = dc.field(default_factory=dict)
    handlers: ta.Dict[str, HandlerDictLoggingConfig] = dc.field(default_factory=dict)
    loggers: ta.Dict[str, LoggerDictLoggingConfig] = dc.field(default_factory=dict)
    root: ta.Optional[LoggerDictLoggingConfig] = None



import typing as ta

from .. import check
from .. import collections as col
from .. import dataclasses as dc
from .. import lang
from .base import AnyLifecycle
from .base import AsyncLifecycle
from .base import Lifecycle
from .base import as_async_lifecycle
from .controller import AnyLifecycleController
from .controller import AsyncLifecycleController
from .controller import LifecycleController
from .managed import AsyncLifecycleManaged
from .managed import LifecycleManaged
from .states import LifecycleState
from .states import LifecycleStateError
from .states import LifecycleStates


LifecycleControllerT = ta.TypeVar('LifecycleControllerT', bound=AnyLifecycleController)


##


@dc.dataclass(frozen=True, eq=False)
class LifecycleManagerEntry(lang.Final, ta.Generic[LifecycleControllerT]):
    controller: LifecycleControllerT

    dependencies: ta.MutableSet['LifecycleManagerEntry'] = dc.field(default_factory=set)
    dependents: ta.MutableSet['LifecycleManagerEntry'] = dc.field(default_factory=set)


#


@ta.final
class _InnerLifecycleManager(AsyncLifecycleManaged, lang.Final):
    def __init__(
            self,
            get_state: ta.Callable[[], LifecycleState],
    ) -> None:
        super().__init__()

        self._get_state = get_state

        self._entries_by_lifecycle: ta.MutableMapping[AnyLifecycle, LifecycleManagerEntry] = col.IdentityKeyDict()

    #

    def _get_or_make_controller(self, lifecycle: AnyLifecycle) -> AnyLifecycleController:
        if isinstance(lifecycle, (LifecycleController, AsyncLifecycleController)):
            return lifecycle

        elif isinstance(lifecycle, Lifecycle):
            return LifecycleController(lifecycle)

        elif isinstance(lifecycle, AsyncLifecycle):
            return AsyncLifecycleController(lifecycle)

        else:
            raise TypeError(lifecycle)

    def _add_internal(self, lifecycle: AnyLifecycle, dependencies: ta.Iterable[AnyLifecycle]) -> LifecycleManagerEntry:
        check.state(self._get_state() < LifecycleStates.STOPPING and not self._get_state().is_failed)

        check.isinstance(lifecycle, (Lifecycle, AsyncLifecycle))
        try:
            entry = self._entries_by_lifecycle[lifecycle]
        except KeyError:
            controller = self._get_or_make_controller(lifecycle)
            entry = self._entries_by_lifecycle[lifecycle] = LifecycleManagerEntry(controller)

        for dep in dependencies:
            check.isinstance(dep, (Lifecycle, AsyncLifecycle))
            dep_entry = self._add_internal(dep, [])
            entry.dependencies.add(dep_entry)
            dep_entry.dependents.add(entry)

        return entry

    async def add(
            self,
            lifecycle: AnyLifecycle,
            dependencies: ta.Iterable[AnyLifecycle] = (),
    ) -> LifecycleManagerEntry:
        check.state(self._get_state() < LifecycleStates.STOPPING and not self._get_state().is_failed)

        entry = self._add_internal(lifecycle, dependencies)

        if self._get_state() >= LifecycleStates.CONSTRUCTING:
            async def rec(e):
                if e.controller.state < LifecycleStates.CONSTRUCTED:
                    for dep in e.dependencies:
                        await rec(dep)

                    await as_async_lifecycle(e.controller).lifecycle_construct()

            await rec(entry)

        if self._get_state() >= LifecycleStates.STARTING:
            async def rec(e):
                if e.controller.state < LifecycleStates.STARTED:
                    for dep in e.dependencies:
                        await rec(dep)

                    await as_async_lifecycle(e.controller).lifecycle_start()

            await rec(entry)

        return entry

    ##

    @ta.override
    async def _lifecycle_construct(self) -> None:
        async def rec(e: LifecycleManagerEntry) -> None:
            for dep in e.dependencies:
                await rec(dep)

            if e.controller.state.is_failed:
                raise LifecycleStateError(e.controller)

            if e.controller.state < LifecycleStates.CONSTRUCTED:
                await as_async_lifecycle(e.controller).lifecycle_construct()

        for entry in self._entries_by_lifecycle.values():
            await rec(entry)

    @ta.override
    async def _lifecycle_start(self) -> None:
        async def rec(e: LifecycleManagerEntry) -> None:
            for dep in e.dependencies:
                await rec(dep)

            if e.controller.state.is_failed:
                raise LifecycleStateError(e.controller)

            if e.controller.state < LifecycleStates.CONSTRUCTED:
                await as_async_lifecycle(e.controller).lifecycle_construct()

            if e.controller.state < LifecycleStates.STARTED:
                await as_async_lifecycle(e.controller).lifecycle_start()

        for entry in self._entries_by_lifecycle.values():
            await rec(entry)

    @ta.override
    async def _lifecycle_stop(self) -> None:
        async def rec(e: LifecycleManagerEntry) -> None:
            for dep in e.dependents:
                await rec(dep)

            if e.controller.state.is_failed:
                raise LifecycleStateError(e.controller)

            if e.controller.state is LifecycleStates.STARTED:
                await as_async_lifecycle(e.controller).lifecycle_stop()

        for entry in self._entries_by_lifecycle.values():
            await rec(entry)

    @ta.override
    async def _lifecycle_destroy(self) -> None:
        async def rec(e: LifecycleManagerEntry) -> None:
            for dep in e.dependents:
                await rec(dep)

            if e.controller.state < LifecycleStates.DESTROYED:
                await as_async_lifecycle(e.controller).lifecycle_destroy()

        for entry in self._entries_by_lifecycle.values():
            await rec(entry)


##


class LifecycleManager(LifecycleManaged, lang.Final):
    def __init__(
            self,
            *,
            lock: lang.DefaultLockable = None,
    ) -> None:
        super().__init__()

        self._lock = lang.default_lock(lock, None)

        self._inner = _InnerLifecycleManager(lambda: self._inner_controller.state)

        self._inner_controller = AsyncLifecycleController(
            self._inner._lifecycle,  # noqa
            lock=lambda: lang.SyncToAsyncContextManager(self._lock()),
        )

    #

    @property
    def state(self) -> LifecycleState:
        return self._inner_controller.state

    #

    def add(
            self,
            lifecycle: Lifecycle,
            dependencies: ta.Iterable[Lifecycle] = (),
    ) -> LifecycleManagerEntry[LifecycleController]:
        return lang.sync_await(self._inner.add(
            check.isinstance(lifecycle, Lifecycle),
            [check.isinstance(d, Lifecycle) for d in dependencies],
        ))

    #

    @ta.override
    def _lifecycle_construct(self) -> None:
        return lang.sync_await(self._inner_controller.lifecycle_construct())

    @ta.override
    def _lifecycle_start(self) -> None:
        return lang.sync_await(self._inner_controller.lifecycle_start())

    @ta.override
    def _lifecycle_stop(self) -> None:
        return lang.sync_await(self._inner_controller.lifecycle_stop())

    @ta.override
    def _lifecycle_destroy(self) -> None:
        return lang.sync_await(self._inner_controller.lifecycle_destroy())


#


class AsyncLifecycleManager(AsyncLifecycleManaged, lang.Final):
    def __init__(
            self,
            *,
            lock: lang.DefaultAsyncLockable = None,
    ) -> None:
        super().__init__()

        self._lock = lang.default_async_lock(lock, None)

        self._inner = _InnerLifecycleManager(lambda: self._inner_controller.state)

        self._inner_controller = AsyncLifecycleController(
            self._inner._lifecycle,  # noqa
            lock=self._lock,
        )

    #

    @property
    def state(self) -> LifecycleState:
        return self._inner_controller.state

    #

    async def add(
            self,
            lifecycle: AnyLifecycle,
            dependencies: ta.Iterable[AnyLifecycle] = (),
    ) -> LifecycleManagerEntry[AnyLifecycleController]:
        return await self._inner.add(
            lifecycle,
            dependencies,
        )

    #

    @ta.override
    async def _lifecycle_construct(self) -> None:
        return await self._inner_controller.lifecycle_construct()

    @ta.override
    async def _lifecycle_start(self) -> None:
        return await self._inner_controller.lifecycle_start()

    @ta.override
    async def _lifecycle_stop(self) -> None:
        return await self._inner_controller.lifecycle_stop()

    @ta.override
    async def _lifecycle_destroy(self) -> None:
        return await self._inner_controller.lifecycle_destroy()



import dataclasses as dc
import typing as ta

from .... import check
from .... import lang
from ..utils import SealableRegistry
from .base import ProcessingContextItemFactory
from .base import Processor
from .priority import ProcessorPriority


ProcessorT = ta.TypeVar('ProcessorT', bound=Processor)


##


_PROCESSING_CONTEXT_ITEM_FACTORIES: SealableRegistry[type, ProcessingContextItemFactory] = SealableRegistry()


def register_processing_context_item_factory(i_ty):
    def inner(fn):
        _PROCESSING_CONTEXT_ITEM_FACTORIES[i_ty] = fn
        return fn

    return inner


@lang.cached_function
def processing_context_item_factory_for(i_ty: type) -> ProcessingContextItemFactory:
    return _PROCESSING_CONTEXT_ITEM_FACTORIES[i_ty]


@lang.cached_function
def all_processing_context_item_factories() -> ta.Mapping[type, ProcessingContextItemFactory]:
    return dict(_PROCESSING_CONTEXT_ITEM_FACTORIES.items())


##


@dc.dataclass(frozen=True, kw_only=True)
class ProcessorTypeRegistration:
    priority: ProcessorPriority


_PROCESSOR_TYPES: SealableRegistry[type[Processor], ProcessorTypeRegistration] = SealableRegistry()


def register_processor_type(
        *,
        priority: ProcessorPriority,
        **kwargs: ta.Any,
) -> ta.Callable[[type[ProcessorT]], type[ProcessorT]]:
    reg = ProcessorTypeRegistration(
        priority=priority,
        **kwargs,
    )

    def inner(ty):
        check.issubclass(ty, Processor)
        _PROCESSOR_TYPES[ty] = reg
        return ty

    return inner


@lang.cached_function
def all_processor_types() -> ta.Mapping[type, ProcessorTypeRegistration]:
    return dict(_PROCESSOR_TYPES.items())


@lang.cached_function
def ordered_processor_types() -> ta.Sequence[type]:
    return [
        t
        for t, r in sorted(
            all_processor_types().items(),
            key=lambda t_r: t_r[1].priority,
        )
    ]



import typing as ta

from .. import check
from .base import AsyncLifecycle
from .base import Lifecycle


##


class LifecycleListener:
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        try:
            async_lifecycle_listener_cls = AsyncLifecycleListener
        except NameError:
            pass
        else:
            check.not_issubclass(cls, async_lifecycle_listener_cls)

    def on_starting(self, obj: Lifecycle) -> None:
        pass

    def on_started(self, obj: Lifecycle) -> None:
        pass

    def on_stopping(self, obj: Lifecycle) -> None:
        pass

    def on_stopped(self, obj: Lifecycle) -> None:
        pass


class AsyncLifecycleListener:
    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        check.not_issubclass(cls, LifecycleListener)

    async def on_starting(self, obj: AsyncLifecycle) -> None:
        pass

    async def on_started(self, obj: AsyncLifecycle) -> None:
        pass

    async def on_stopping(self, obj: AsyncLifecycle) -> None:
        pass

    async def on_stopped(self, obj: AsyncLifecycle) -> None:
        pass


AnyLifecycleListener: ta.TypeAlias = LifecycleListener | AsyncLifecycleListener

ANY_LIFECYCLE_LISTENER_TYPES: tuple[type[LifecycleListener | AsyncLifecycleListener], ...] = (LifecycleListener, AsyncLifecycleListener)  # noqa



import typing as ta

from ..... import check
from ...._internals import StdParams
from ....specs import ClassSpec
from ....specs import FieldSpec


##


def std_params_to_class_spec(
        p: StdParams,
        fields: ta.Sequence[FieldSpec],
        *,
        metadata: ta.Sequence[ta.Any] | None = None,
) -> ClassSpec:
    return ClassSpec(
        fields=fields,

        init=check.isinstance(p.init, bool),
        repr=check.isinstance(p.repr, bool),
        eq=check.isinstance(p.eq, bool),
        order=check.isinstance(p.order, bool),
        unsafe_hash=check.isinstance(p.unsafe_hash, bool),
        frozen=check.isinstance(p.frozen, bool),

        match_args=check.isinstance(p.match_args, bool),
        kw_only=check.isinstance(p.kw_only, bool),
        slots=check.isinstance(p.slots, bool),
        weakref_slot=check.isinstance(p.weakref_slot, bool),

        metadata=metadata,
    )



import abc
import typing as ta

from .. import check
from .. import lang
from .elements import CollectedElements
from .elements import Elemental
from .elements import as_elements
from .elements import collect_elements
from .inspect import KwargsTarget
from .keys import Key


if ta.TYPE_CHECKING:
    from .impl import injector as _injector
else:
    _injector = lang.proxy_import('.impl.injector', __package__)


T = ta.TypeVar('T')


##


class AsyncInjector(lang.Abstract):
    @abc.abstractmethod
    def try_provide(self, key: ta.Any) -> ta.Awaitable[lang.Maybe[ta.Any]]:
        raise NotImplementedError

    @abc.abstractmethod
    def provide(self, key: ta.Any) -> ta.Awaitable[ta.Any]:
        raise NotImplementedError

    @abc.abstractmethod
    def provide_kwargs(self, kt: KwargsTarget) -> ta.Awaitable[ta.Mapping[str, ta.Any]]:
        raise NotImplementedError

    @abc.abstractmethod
    def inject(self, obj: ta.Any) -> ta.Awaitable[ta.Any]:
        raise NotImplementedError

    def __getitem__(
            self,
            target: Key[T] | type[T],
    ) -> ta.Awaitable[T]:
        return self.provide(target)


##


@ta.final
class _InjectorCreator(ta.Generic[T]):
    def __init__(self, fac: ta.Callable[[CollectedElements], T]) -> None:
        self._fac = fac

    @ta.overload
    def __call__(self, es: CollectedElements, /) -> T: ...

    @ta.overload
    def __call__(self, *es: Elemental) -> T: ...

    def __call__(self, arg0, *argv):
        ce: CollectedElements
        if isinstance(arg0, CollectedElements):
            check.arg(not argv)
            ce = arg0
        else:
            ce = collect_elements(as_elements(arg0, *argv))
        return self._fac(ce)


##


create_async_injector = _InjectorCreator[ta.Awaitable[AsyncInjector]](lambda ce: _injector.create_async_injector(ce))



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import os
import re
import sys
import typing as ta


##


# Set by pyproject, docker-dev script
DOCKER_HOST_PLATFORM_KEY = 'DOCKER_HOST_PLATFORM'


def get_docker_host_platform() -> ta.Optional[str]:
    return os.environ.get(DOCKER_HOST_PLATFORM_KEY)


##


_LIKELY_IN_DOCKER_PATTERN = re.compile(r'^overlay / .*/(docker|desktop-containerd)/')


def is_likely_in_docker() -> bool:
    if getattr(sys, 'platform') != 'linux':
        return False
    with open('/proc/mounts') as f:
        ls = f.readlines()
    return any(_LIKELY_IN_DOCKER_PATTERN.match(l) for l in ls)



import typing as ta

from ... import lang
from .registries import Registry
from .registries import RegistryItem
from .registries import RegistryView


ConfigT = ta.TypeVar('ConfigT', bound='Config')


##


class Config(RegistryItem, lang.Abstract):
    pass


Configs: ta.TypeAlias = RegistryView[Config]


##


ConfigRegistry: ta.TypeAlias = Registry[Config]

lang.static_check_issubclass[Configs](ConfigRegistry)


EMPTY_CONFIG_REGISTRY = ConfigRegistry().seal()



"""
TODO:
 - ensure all 'init' fields work - non-instance
 - special case 'None' default, most common
"""
import dataclasses as dc
import itertools
import typing as ta

from .... import check
from ..._internals import STD_POST_INIT_NAME
from ...inspect import FieldsInspection
from ...specs import CoerceFn
from ...specs import DefaultFactory
from ...specs import FieldSpec
from ...specs import FieldType
from ...specs import InitFn
from ...specs import ValidateFn
from ..generation.base import Generator
from ..generation.base import Plan
from ..generation.base import PlanResult
from ..generation.globals import FIELD_FN_VALIDATION_ERROR_GLOBAL
from ..generation.globals import FIELD_TYPE_VALIDATION_ERROR_GLOBAL
from ..generation.globals import FN_VALIDATION_ERROR_GLOBAL
from ..generation.globals import HAS_DEFAULT_FACTORY_GLOBAL
from ..generation.globals import ISINSTANCE_GLOBAL
from ..generation.globals import NONE_GLOBAL
from ..generation.idents import SELF_IDENT
from ..generation.ops import AddMethodOp
from ..generation.ops import Op
from ..generation.ops import OpRef
from ..generation.ops import Ref
from ..generation.registry import register_generator_type
from ..generation.utils import SetattrSrcBuilder
from ..processing.base import ProcessingContext
from .fields import InitFields
from .mro import MroDict


##


@dc.dataclass(frozen=True, kw_only=True)
class InitPlan(Plan):
    @dc.dataclass(frozen=True)
    class Field:
        name: str
        annotation: OpRef[ta.Any]

        default: OpRef[ta.Any] | None
        default_factory: OpRef[ta.Any] | None

        init: bool

        override: bool

        field_type: FieldType

        coerce: bool | OpRef[CoerceFn] | None
        validate: OpRef[ValidateFn] | None

        check_type: OpRef[type | tuple[type, ...]] | None

    fields: tuple[Field, ...]

    self_param: str
    std_params: tuple[str, ...]
    kw_only_params: tuple[str, ...]

    frozen: bool

    slots: bool

    post_init_params: tuple[str, ...] | None

    init_fns: tuple[OpRef[InitFn], ...]

    @dc.dataclass(frozen=True)
    class ValidateFnWithParams:
        fn: OpRef[ValidateFn]
        params: tuple[str, ...]

    validate_fns: ta.Sequence[ValidateFnWithParams] | None


@register_generator_type(InitPlan)
class InitGenerator(Generator[InitPlan]):
    def _plan_field(
            self,
            ctx: ProcessingContext,
            i: int,
            f: FieldSpec,
            ann: ta.Any,
            ref_map: dict,
    ) -> InitPlan.Field:
        ann_ref: OpRef = OpRef(f'init.fields.{i}.annotation')
        ref_map[ann_ref] = ann

        default_ref: OpRef[ta.Any] | None = None
        default_factory_ref: OpRef[ta.Any] | None = None
        if f.default.present:
            dfl = f.default.must()
            if isinstance(dfl, DefaultFactory):
                default_factory_ref = OpRef(f'init.fields.{i}.default_factory')
                ref_map[default_factory_ref] = dfl.fn
            else:
                default_ref = OpRef(f'init.fields.{i}.default')
                ref_map[default_ref] = dfl

        coerce: bool | OpRef[CoerceFn] | None = None
        if isinstance(f.coerce, bool):
            coerce = f.coerce
        elif f.coerce is not None:
            coerce = OpRef(f'init.fields.{i}.coerce')
            ref_map[coerce] = f.coerce

        validate_ref: OpRef[ValidateFn] | None = None
        if f.validate is not None:
            validate_ref = OpRef(f'init.fields.{i}.validate')
            ref_map[validate_ref] = f.validate

        check_type_ref: OpRef[type | tuple[type, ...]] | None = None
        if f.check_type is not None and f.check_type is not False:
            check_type_arg: ta.Any
            if isinstance(f.check_type, tuple):
                check_type_arg = tuple(type(None) if e is None else check.isinstance(e, type) for e in f.check_type)
            elif isinstance(f.check_type, type):
                check_type_arg = f.check_type
            elif f.check_type is True:
                check_type_arg = f.annotation
            else:
                raise TypeError(f.check_type)
            check_type_ref = OpRef(f'init.fields.{i}.check_type')
            ref_map[check_type_ref] = check_type_arg

        return InitPlan.Field(
            name=f.name,
            annotation=ann_ref,

            default=default_ref,
            default_factory=default_factory_ref,

            init=f.init,

            override=f.override or ctx.cs.override,

            field_type=f.field_type,

            coerce=coerce,
            validate=validate_ref,

            check_type=check_type_ref,
        )

    def plan(self, ctx: ProcessingContext) -> PlanResult[InitPlan] | None:
        if '__init__' in ctx.cls.__dict__ or not ctx.cs.init:
            return None

        init_fields = ctx[InitFields]
        seen_default = None
        for f in init_fields.std:
            if not f.init:
                continue
            if f.default.present:
                seen_default = f
            elif seen_default:
                raise TypeError(f'non-default argument {f.name!r} follows default argument {seen_default.name!r}')

        if ctx.cs.generic_init:
            gr_field_anns = ctx[FieldsInspection].generic_replaced_field_annotations
            get_field_ann = lambda f: gr_field_anns[f.name]
        else:
            get_field_ann = lambda f: f.annotation

        ref_map: dict = {}

        plan_fields: list[InitPlan.Field] = []
        for i, f in enumerate(ctx.cs.fields):
            plan_fields.append(self._plan_field(
                ctx,
                i,
                f,
                get_field_ann(f),
                ref_map,
            ))

        mro_v_ids = set(map(id, ctx[MroDict].values()))
        props_by_fget_id = {
            id(v.fget): v
            for v in ctx[MroDict].values()
            if isinstance(v, property)
            and v.fget is not None
        }
        init_fns: list[OpRef[InitFn]] = []
        for i, init_fn in enumerate(ctx.cs.init_fns or []):
            if (obj_id := id(init_fn)) not in mro_v_ids and obj_id in props_by_fget_id:
                init_fn = props_by_fget_id[obj_id].__get__
            elif isinstance(init_fn, property):
                init_fn = init_fn.__get__
            init_fn_ref: OpRef = OpRef(f'init.init_fns.{i}')
            ref_map[init_fn_ref] = init_fn
            init_fns.append(init_fn_ref)

        validate_fns: list[InitPlan.ValidateFnWithParams] = []
        for i, validate_fn in enumerate(ctx.cs.validate_fns or []):
            validate_fn_ref: OpRef = OpRef(f'init.validate_fns.{i}')
            ref_map[validate_fn_ref] = validate_fn.fn
            validate_fns.append(InitPlan.ValidateFnWithParams(
                fn=validate_fn_ref,
                params=tuple(validate_fn.params),
            ))

        post_init_params: tuple[str, ...] | None = None
        if hasattr(ctx.cls, STD_POST_INIT_NAME):
            post_init_params = tuple(f.name for f in init_fields.all if f.field_type is FieldType.INIT_VAR)

        return PlanResult(
            InitPlan(
                fields=tuple(plan_fields),

                self_param=SELF_IDENT if 'self' in ctx.cs.fields_by_name else 'self',
                std_params=tuple(f.name for f in init_fields.std),
                kw_only_params=tuple(f.name for f in init_fields.kw_only),

                frozen=ctx.cs.frozen,

                slots=ctx.cs.slots,

                post_init_params=post_init_params,

                init_fns=tuple(init_fns),

                validate_fns=tuple(validate_fns),
            ),
            ref_map,
        )

    def generate(self, plan: InitPlan) -> ta.Iterable[Op]:
        refs: set[Ref] = set()

        fields_by_name = {f.name: f for f in plan.fields}

        # proto

        params: list[str] = []
        seen_kw_only = False
        for fn, kw_only in itertools.chain(
            [(fn, False) for fn in plan.std_params],
            [(fn, True) for fn in plan.kw_only_params],
        ):
            f = fields_by_name[fn]
            if kw_only:
                if not seen_kw_only:
                    params.append('*')
                    seen_kw_only = True
            elif seen_kw_only:
                raise TypeError(f'non-keyword-only argument {f.name!r} follows keyword-only argument(s)')

            refs.add(f.annotation)
            p = f'{f.name}: {f.annotation.ident()}'

            if f.default_factory is not None:
                check.none(f.default)
                p += f' = {HAS_DEFAULT_FACTORY_GLOBAL.ident}'
                refs.add(HAS_DEFAULT_FACTORY_GLOBAL)
            elif f.default is not None:
                check.none(f.default_factory)
                refs.add(f.default)
                p += f' = {f.default.ident()}'

            params.append(p)

        proto_lines = [
            f'def __init__(',
            f'    {plan.self_param},',
            *[
                f'    {p},'
                for p in params
            ],
            f') -> {NONE_GLOBAL.ident}:',
        ]
        refs.add(NONE_GLOBAL)

        # body

        lines = []

        # defaults

        values: dict[str, str] = {
            plan.self_param: plan.self_param,
        }

        for f in plan.fields:
            if f.default_factory is not None:
                check.none(f.default)
                refs.add(f.default_factory)
                if f.init:
                    lines.extend([
                        f'    if {f.name} is {HAS_DEFAULT_FACTORY_GLOBAL.ident}:',
                        f'        {f.name} = {f.default_factory.ident()}()',
                    ])
                    refs.add(HAS_DEFAULT_FACTORY_GLOBAL)
                else:
                    lines.append(
                        f'    {f.name} = {f.default_factory.ident()}()',
                    )
                values[f.name] = f.name

            elif f.init:
                if f.default is not None:
                    check.none(f.default_factory)
                    values[f.name] = f.name

                else:
                    values[f.name] = f.name

            elif plan.slots and f.default is not None:
                refs.add(f.default)
                lines.append(
                    f'    {f.name} = {f.default.ident()}',
                )
                values[f.name] = f.name

        # coercion

        for f in plan.fields:
            if isinstance(f.coerce, bool) and f.coerce:
                lines.append(
                    f'    {f.name} = {f.annotation.ident()}({values[f.name]})',
                )
                values[f.name] = f.name
            elif isinstance(f.coerce, OpRef):
                refs.add(f.coerce)
                lines.append(
                    f'    {f.name} = {f.coerce.ident()}({values[f.name]})',
                )
                values[f.name] = f.name

        # field validation

        for f in plan.fields:
            if f.check_type is None:
                continue
            refs.add(f.check_type)
            lines.extend([
                f'    if not {ISINSTANCE_GLOBAL.ident}({values[f.name]}, {f.check_type.ident()}): ',
                f'        raise {FIELD_TYPE_VALIDATION_ERROR_GLOBAL.ident}(',
                f'            obj={plan.self_param},',
                f'            type={f.check_type.ident()},',
                f'            field={f.name!r},',
                f'            value={values[f.name]},',
                f'        )',
            ])
            refs.add(ISINSTANCE_GLOBAL)
            refs.add(FIELD_TYPE_VALIDATION_ERROR_GLOBAL)

        for f in plan.fields:
            if f.validate is None:
                continue
            refs.add(f.validate)
            lines.extend([
                f'    if not {f.validate.ident()}({values[f.name]}): ',
                f'        raise {FIELD_FN_VALIDATION_ERROR_GLOBAL.ident}(',
                f'            obj={plan.self_param},',
                f'            fn={f.validate.ident()},',
                f'            field={f.name!r},',
                f'            value={values[f.name]},',
                f'        )',
            ])
            refs.add(FIELD_FN_VALIDATION_ERROR_GLOBAL)

        # setattr

        sab = SetattrSrcBuilder(
            object_ident=plan.self_param,
        )
        for f in plan.fields:
            if f.name not in values or f.field_type != FieldType.INSTANCE:
                continue
            lines.extend([
                f'    {l}'
                for l in sab(
                    f.name,
                    values[f.name],
                    frozen=plan.frozen,
                    override=f.override,
                )
            ])
        refs.update(sab.refs)

        # fn validation

        for vfn in plan.validate_fns or []:
            refs.add(vfn.fn)
            if vfn.params:
                lines.extend([
                    f'    if not {vfn.fn.ident()}(',
                    *[
                        f'        {values[p]},'
                        for p in vfn.params
                    ],
                    f'    ):',
                ])
            else:
                lines.append(
                    f'    if not {vfn.fn.ident()}():',
                )
            lines.extend([
                f'        raise {FN_VALIDATION_ERROR_GLOBAL.ident}(',
                f'            obj={plan.self_param},',
                f'            fn={vfn.fn.ident()},',
                f'        )',
            ])
            refs.add(FN_VALIDATION_ERROR_GLOBAL)

        # post-init

        if (pia := plan.post_init_params) is not None:
            if pia:
                lines.extend([
                    f'    {plan.self_param}.{STD_POST_INIT_NAME}(',
                    *[
                        f'        {values[p]},'
                        for p in pia
                    ],
                    f'    )',
                ])
            else:
                lines.append(
                    f'    {plan.self_param}.{STD_POST_INIT_NAME}()',
                )

        for init_fn in plan.init_fns:
            refs.add(init_fn)
            lines.append(
                f'    {init_fn.ident()}({plan.self_param})',
            )

        #

        if not lines:
            lines.append(
                '    pass',
            )

        return [
            AddMethodOp(
                '__init__',
                '\n'.join([*proto_lines, *lines]),
                frozenset(refs),
            ),
        ]



import io
import re
import typing as ta

from ... import check
from ... import lang
from ..json import Scalar
from ..json.literals import ESCAPE_MAP
from ..json.literals import encode_string
from ..json.rendering import JsonRenderer
from ..json.rendering import JsonRendererOut


##


MULTILINE_STRINGS_ENDL = '\\\n'
MULTILINE_STRINGS_LQ = '"' + MULTILINE_STRINGS_ENDL
MULTILINE_STRINGS_RQ = MULTILINE_STRINGS_ENDL + '"'
MULTILINE_STRINGS_NL = '\\n' + MULTILINE_STRINGS_ENDL

MULTILINE_STRINGS_ESCAPE_MAP = {
    **ESCAPE_MAP,
    '\n': MULTILINE_STRINGS_NL,
}

SOFTWRAP_WS_PAT = re.compile(r'\s+')


class Json5Renderer(JsonRenderer):
    def __init__(
            self,
            out: JsonRendererOut,
            *,
            multiline_strings: bool = False,
            softwrap_length: int | None = None,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(out, **kwargs)

        self._multiline_strings = multiline_strings
        self._softwrap_length = softwrap_length

    def _softwrap_string_chunks(self, chunks: list[str]) -> str:
        multiline_strings = self._multiline_strings
        softwrap_len = check.not_none(self._softwrap_length)

        out = io.StringIO()
        out.write(MULTILINE_STRINGS_LQ)

        l = 0

        def write(s: str) -> None:
            nonlocal l
            if l >= (softwrap_len - (sl := len(s))):
                out.write(MULTILINE_STRINGS_ENDL)
                l = 0

            out.write(s)
            l += sl

        for c in chunks:
            if not c:
                continue

            if c == '\\n' and multiline_strings:
                if (softwrap_len - l) > 2:
                    out.write('\\n\\\n')
                else:
                    out.write('\\\n\\n\\\n')
                l = 0
                continue

            it: ta.Iterable[str | re.Match]
            if c[0] == '\\':
                it = [c]
            else:
                it = lang.iter_pat(SOFTWRAP_WS_PAT, c)

            for x in it:
                if isinstance(x, re.Match):
                    ws = x.group(0)
                    r = softwrap_len - l
                    if len(ws) > r:
                        write(ws[:r])
                        p = r
                        while p < len(ws):
                            write(ws[p:(np := (p + softwrap_len))])
                            p = np

                    else:
                        write(ws)

                else:
                    write(x)

        out.write(MULTILINE_STRINGS_RQ)
        return out.getvalue()

    def _format_string(self, s: str, state: JsonRenderer.State | None = None) -> str:
        num_nls = s.count('\n')
        is_multiline = self._multiline_strings and num_nls

        if (softwrap_len := self._softwrap_length) is not None:
            def process_chunks(chunks: list[str]) -> list[str]:
                naive_len = sum(map(len, chunks))

                if is_multiline:
                    if naive_len < (
                        softwrap_len -
                        len(MULTILINE_STRINGS_LQ) -
                        (num_nls * (len(MULTILINE_STRINGS_NL) - 1)) -
                        len(MULTILINE_STRINGS_RQ)
                    ):
                        return [
                            MULTILINE_STRINGS_LQ,
                            *[
                                MULTILINE_STRINGS_NL if c == '\\n' else c
                                for c in chunks
                            ],
                            MULTILINE_STRINGS_RQ,
                        ]

                elif naive_len < softwrap_len:
                    return [
                        '"',
                        *chunks,
                        '"',
                    ]

                return [self._softwrap_string_chunks(chunks)]

            return encode_string(
                s,
                q='',
                ensure_ascii=self._ensure_ascii,
                process_chunks=process_chunks,
            )

        if is_multiline:
            return encode_string(
                s,
                lq=MULTILINE_STRINGS_LQ,
                rq=MULTILINE_STRINGS_RQ,
                escape_map=MULTILINE_STRINGS_ESCAPE_MAP,
                ensure_ascii=self._ensure_ascii,
            )

        return super()._format_scalar(s, state=state)

    def _format_scalar(self, o: Scalar, state: JsonRenderer.State | None = None) -> str:
        if isinstance(o, str):
            return self._format_string(o)

        else:
            return super()._format_scalar(o, state=state)



import typing as ta

from ..lite.maysync import MaysyncFn
from ..lite.maysync import MaysyncGeneratorFn
from ..lite.maysync import make_maysync as _make_maysync
from .asyncs import as_async


T = ta.TypeVar('T')
O = ta.TypeVar('O')
I = ta.TypeVar('I')

P = ta.ParamSpec('P')


##


def make_maysync_fn(
        s: ta.Callable[P, T],
        a: ta.Callable[P, ta.Awaitable[T]],
) -> ta.Callable[P, ta.Awaitable[T]]:
    """Constructs a MaysyncFn from a (sync, async) function pair."""

    return MaysyncFn(s, a)


def make_maysync_generator_fn(
        s: ta.Callable[P, ta.Generator[O, I]],
        a: ta.Callable[P, ta.AsyncGenerator[O, I]],
) -> ta.Callable[P, ta.AsyncGenerator[O, I]]:
    """Constructs a MaysyncGeneratorFn from a (sync, async) generator function pair."""

    return MaysyncGeneratorFn(s, a)


@ta.overload
def make_maysync(
        s: ta.Callable[P, T],
        a: ta.Callable[P, ta.Awaitable[T]],
) -> ta.Callable[P, ta.Awaitable[T]]:
    ...


@ta.overload
def make_maysync(
        s: ta.Callable[P, ta.Generator[O, I]],
        a: ta.Callable[P, ta.AsyncGenerator[O, I]],
) -> ta.Callable[P, ta.AsyncGenerator[O, I]]:
    ...


def make_maysync(s, a):
    """
    Constructs a MaysyncFn or MaysyncGeneratorFn from a (sync, async) function pair, using `inspect.isasyncgenfunction`
    to determine the type.
    """

    return _make_maysync(s, a)


##


def make_maysync_from_sync(
        s: ta.Callable[P, T],
        a: ta.Callable[P, ta.Awaitable[T]] | None = None,
) -> ta.Callable[P, ta.Awaitable[T]]:
    return _make_maysync(
        s,
        a if a is not None else as_async(s, wrap=True),
    )



from .asquery import (  # noqa
    AsQueryParams,
    as_query,
    as_query_,
)

from .base import (  # noqa
    Querier,
    Rows,
    Conn,
    Db,
    Adapter,
)

from .columns import (  # noqa
    Column,
    Columns,
)

from .dbapi import (  # noqa
    DbapiRows,
    DbapiConn,
    DbapiDb,
    DbapiAdapter,
)

from .errors import (  # noqa
    Error,

    ColumnError,
    DuplicateColumnNameError,
    MismatchedColumnCountError,

    QueryError,
)

from .funcs import (  # noqa
    exec,  # noqa

    query,
    query_all,
    query_first,
    query_opt_first,
    query_one,
    query_opt_one,
    query_scalar,
    query_maybe_scalar,
)

from .queries import (  # noqa
    QueryMode,
    Query,
)

from .resources import (  # noqa
    get_resource_debug,
    set_resource_debug,

    UnclosedResourceWarning,
    Closer,

    ResourceNotEnteredError,
    ContextCloser,
)

from .rows import (  # noqa
    Row,
)


##


from ... import lang as _lang

_lang.register_conditional_import('..queries', '._queries', __package__)



import typing as ta

from .. import check
from .. import lang
from .base import AsyncLifecycle
from .base import Lifecycle
from .listeners import AsyncLifecycleListener
from .listeners import LifecycleListener
from .states import LifecycleState
from .states import LifecycleStates
from .transitions import LifecycleTransition
from .transitions import LifecycleTransitions


##


@ta.final
class LifecycleController(Lifecycle, lang.Final):
    def __init__(
            self,
            controlled: Lifecycle,
            *,
            lock: lang.DefaultLockable = None,
    ) -> None:
        super().__init__()

        self._controlled: Lifecycle = check.isinstance(controlled, Lifecycle)
        self._lock = lang.default_lock(lock, None)

        self._state = LifecycleStates.NEW
        self._listeners: list[LifecycleListener] = []

    __repr__ = lang.attr_ops(lambda o: (o.controlled, o.state)).repr

    #

    @property
    def controlled(self) -> Lifecycle:
        return self._controlled

    @property
    def state(self) -> LifecycleState:
        return self._state

    def add_listener(self, listener: LifecycleListener) -> ta.Self:
        self._listeners.append(check.isinstance(listener, LifecycleListener))
        return self

    #

    def _set_state(self, state: LifecycleState) -> None:
        self._state = state
        self._controlled.lifecycle_state(state)

    def _advance(
            self,
            transition: LifecycleTransition,
            controlled_fn: ta.Callable[[], None],
            pre_listener_fn: ta.Callable[[LifecycleListener], ta.Callable[[Lifecycle], None]] | None = None,
            post_listener_fn: ta.Callable[[LifecycleListener], ta.Callable[[Lifecycle], None]] | None = None,
    ) -> None:
        with self._lock():
            if pre_listener_fn is not None:
                for listener in self._listeners:
                    pre_listener_fn(listener)(self._controlled)

            check.state(self._state in transition.old)
            self._set_state(transition.new_intermediate)

            try:
                controlled_fn()
            except Exception:
                self._set_state(transition.new_failed)
                raise

            self._set_state(transition.new_succeeded)

            if post_listener_fn is not None:
                for listener in self._listeners:
                    post_listener_fn(listener)(self._controlled)

    ##

    @ta.override
    def lifecycle_construct(self) -> None:
        self._advance(
            LifecycleTransitions.CONSTRUCT,
            self._controlled.lifecycle_construct,
        )

    @ta.override
    def lifecycle_start(self) -> None:
        self._advance(
            LifecycleTransitions.START,
            self._controlled.lifecycle_start,
            lambda l: l.on_starting,
            lambda l: l.on_started,
        )

    @ta.override
    def lifecycle_stop(self) -> None:
        self._advance(
            LifecycleTransitions.STOP,
            self._controlled.lifecycle_stop,
            lambda l: l.on_stopping,
            lambda l: l.on_stopped,
        )

    @ta.override
    def lifecycle_destroy(self) -> None:
        self._advance(
            LifecycleTransitions.DESTROY,
            self._controlled.lifecycle_destroy,
        )


#


@ta.final
class AsyncLifecycleController(AsyncLifecycle, lang.Final):
    def __init__(
            self,
            controlled: AsyncLifecycle,
            *,
            lock: lang.DefaultAsyncLockable = None,
    ) -> None:
        super().__init__()

        self._controlled: AsyncLifecycle = check.isinstance(controlled, AsyncLifecycle)
        self._lock = lang.default_async_lock(lock, None)

        self._state = LifecycleStates.NEW
        self._listeners: list[AsyncLifecycleListener] = []

    __repr__ = lang.attr_ops(lambda o: (o.controlled, o.state)).repr

    #

    @property
    def controlled(self) -> AsyncLifecycle:
        return self._controlled

    @property
    def state(self) -> LifecycleState:
        return self._state

    def add_listener(self, listener: AsyncLifecycleListener) -> ta.Self:
        self._listeners.append(check.isinstance(listener, AsyncLifecycleListener))
        return self

    #

    async def _set_state(self, state: LifecycleState) -> None:
        self._state = state
        await self._controlled.lifecycle_state(state)

    async def _advance(
            self,
            transition: LifecycleTransition,
            controlled_fn: ta.Callable[[], ta.Awaitable[None]],
            pre_listener_fn: ta.Callable[[AsyncLifecycleListener], ta.Callable[[AsyncLifecycle], ta.Awaitable[None]]] | None = None,  # noqa
            post_listener_fn: ta.Callable[[AsyncLifecycleListener], ta.Callable[[AsyncLifecycle], ta.Awaitable[None]]] | None = None,  # noqa
    ) -> None:
        async with self._lock():
            if pre_listener_fn is not None:
                for listener in self._listeners:
                    await pre_listener_fn(listener)(self._controlled)

            check.state(self._state in transition.old)
            await self._set_state(transition.new_intermediate)

            try:
                await controlled_fn()
            except Exception:
                await self._set_state(transition.new_failed)
                raise

            await self._set_state(transition.new_succeeded)

            if post_listener_fn is not None:
                for listener in self._listeners:
                    await post_listener_fn(listener)(self._controlled)

    ##

    @ta.override
    async def lifecycle_construct(self) -> None:
        await self._advance(
            LifecycleTransitions.CONSTRUCT,
            self._controlled.lifecycle_construct,
        )

    @ta.override
    async def lifecycle_start(self) -> None:
        await self._advance(
            LifecycleTransitions.START,
            self._controlled.lifecycle_start,
            lambda l: l.on_starting,
            lambda l: l.on_started,
        )

    @ta.override
    async def lifecycle_stop(self) -> None:
        await self._advance(
            LifecycleTransitions.STOP,
            self._controlled.lifecycle_stop,
            lambda l: l.on_stopping,
            lambda l: l.on_stopped,
        )

    @ta.override
    async def lifecycle_destroy(self) -> None:
        await self._advance(
            LifecycleTransitions.DESTROY,
            self._controlled.lifecycle_destroy,
        )


#


AnyLifecycleController: ta.TypeAlias = LifecycleController | AsyncLifecycleController

ANY_LIFECYCLE_CONTROLLER_TYPES: tuple[type[LifecycleController | AsyncLifecycleController], ...] = (LifecycleController, AsyncLifecycleController)  # noqa



import abc
import os
import signal
import sys
import time
import typing as ta

from ... import lang


##


class Deathpact(lang.Abstract):
    @abc.abstractmethod
    def poll(self) -> None:
        raise NotImplementedError


class NopDeathpact(Deathpact):
    def poll(self) -> None:
        pass


##


class BaseDeathpact(Deathpact, lang.Abstract):
    def __init__(
            self,
            *,
            interval_s: float = .5,
            signal: int | None = signal.SIGTERM,  # noqa
            output: ta.Literal['stdout', 'stderr'] | None = 'stderr',
            on_die: ta.Callable[[], None] | None = None,
    ) -> None:
        super().__init__()

        self._interval_s = interval_s
        self._signal = signal
        self._output = output
        self._on_die = on_die

        self._last_check_t: float | None = None

    def _print(self, msg: str) -> None:
        match self._output:
            case 'stdout':
                f = sys.stdout
            case 'stderr':
                f = sys.stderr
            case _:
                return
        print(f'{self} pid={os.getpid()}: {msg}', file=f)

    def die(self) -> None:
        self._print('Triggered! Process terminating!')

        if self._on_die is not None:
            self._on_die()

        if self._signal is not None:
            os.kill(os.getpid(), self._signal)

        sys.exit(1)

    @abc.abstractmethod
    def should_die(self) -> bool:
        raise NotImplementedError

    def maybe_die(self) -> None:
        if self.should_die():
            self.die()

    def poll(self) -> None:
        if self._last_check_t is None or (time.monotonic() - self._last_check_t) >= self._interval_s:
            self.maybe_die()
            self._last_check_t = time.monotonic()



import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang


##


@dc.dataclass(frozen=True)
@dc.extra_class_params(terse_repr=True)
class Part(lang.Abstract, lang.Sealed):
    pass


@dc.dataclass(frozen=True)
@dc.extra_class_params(terse_repr=True)
class Text(Part, lang.Final):
    s: str

    @dc.init
    def _check_s(self) -> None:
        check.non_empty_str(self.s)
        check.state(self.s == self.s.strip())


@dc.dataclass(frozen=True)
@dc.extra_class_params(terse_repr=True)
class Blank(Part, lang.Final):
    pass


@dc.dataclass(frozen=True)
@dc.extra_class_params(terse_repr=True)
class Indent(Part, lang.Final):
    n: int = dc.xfield(validate=lambda n: n > 0)
    p: ta.Union[Text, 'Block', 'List'] = dc.xfield(coerce=lambda p: check.isinstance(p, (Text, Block, List)))


@dc.dataclass(frozen=True)
@dc.extra_class_params(terse_repr=True)
class Block(Part, lang.Final):
    ps: ta.Sequence[Part]

    @dc.init
    def _check_ps(self) -> None:
        check.state(len(self.ps) > 1)
        for i, p in enumerate(self.ps):
            check.isinstance(p, Part)
            if i and isinstance(p, Block):
                check.not_isinstance(self.ps[i - 1], Block)


@dc.dataclass(frozen=True)
@dc.extra_class_params(terse_repr=True)
class List(Part, lang.Final):
    d: str = dc.xfield(coerce=check.non_empty_str)
    es: ta.Sequence[Part] = dc.xfield()

    @dc.init
    def _check_es(self) -> None:
        check.not_empty(self.es)
        for e in self.es:
            check.isinstance(e, Part)


##


def _squish(ps: ta.Sequence[Part]) -> ta.Sequence[Part]:
    for p in ps:
        check.isinstance(p, Part)

    if len(ps) < 2:
        return ps

    while True:
        if any(isinstance(p, Block) for p in ps):
            ps = list(lang.flatmap(lambda p: p.ps if isinstance(p, Block) else [p], ps))
            continue

        if any(
                isinstance(ps[i], Indent) and
                isinstance(ps[i + 1], Indent) and
                ps[i].n == ps[i + 1].n  # type: ignore[attr-defined]
                for i in range(len(ps) - 1)
        ):
            new: list[Part | tuple[int, list[Part]]] = []
            for p in ps:
                if isinstance(p, Indent):
                    if new and isinstance(y := new[-1], tuple) and p.n == y[0]:
                        y[1].append(p.p)
                    else:
                        new.append((p.n, [p.p]))
                else:
                    new.append(p)
            ps = [
                Indent(x[0], blockify(*x[1])) if isinstance(x, tuple) else x  # type: ignore[arg-type]
                for x in new
            ]
            continue

        break

    return ps


def blockify(*ps: Part) -> Part:
    check.not_empty(ps)
    ps = _squish(ps)  # type: ignore[assignment]
    if len(ps) == 1:
        return ps[0]
    return Block(ps)


def unblockify(p: Part) -> ta.Sequence[Part]:
    if isinstance(p, Block):
        return p.ps
    else:
        return [p]


##


def build_root(s: str) -> Part:
    lst: list[Part] = []

    for l in s.split('\n'):
        if not (sl := l.strip()):
            lst.append(Blank())
            continue

        p: Part = Text(sl)

        n = next((i for i, c in enumerate(l) if not c.isspace()), 0)
        if n:
            p = Indent(n, p)  # type: ignore[arg-type]

        lst.append(p)

    if len(lst) == 1:
        return lst[0]
    else:
        return Block(lst)
